import math
import numpy

import OpenGL
from OpenGL.GL import *

from PyQt4 import QtGui
from PyQt4.QtCore import *
from PyQt4.QtOpenGL import *


import pomsets_app.gui.canvas as CanvasModule
import pomsets_app.gui.graph as GraphModule




class Canvas(QGLWidget, CanvasModule.GlCanvas):

    ATTRIBUTES = CanvasModule.GlCanvas.ATTRIBUTES + []

    MODE_MOUSE_NONE = 0
    MODE_MOUSE_PAN = 1
    MODE_MOUSE_ZOOM = 2
    MODE_MOUSE_MOVE_OBJECTS = 3
    MODE_MOUSE_SELECTION_BOX = 4
    MODE_MOUSE_CREATE_EDGE = 5

    def __init__(self, parent, width, height, *args, **kwds):
        QGLWidget.__init__(self, parent)
        CanvasModule.GlCanvas.__init__(self, parent, width, height, 
                                       *args, **kwds)
        self.setMinimumSize(600, 400)

        # we want mouse tracking
        # so that we can get tooltips
        self.setMouseTracking(True)

        self.setContextMenuPolicy( Qt.CustomContextMenu )
        self.connect(self, SIGNAL('customContextMenuRequested(const QPoint&)'), self.OnContextMenu)

        self.textureMap = {}

        self.setAcceptDrops(True)
        return


    def setCurrentContextModified(self):
        self.contextManager().currentDisplayedPomsetInfo()[0].isModified(True)
        return


    def mouseMoveMode(self, value=None):
        if value is not None:
            self._mouseMoveMode = value
        if not hasattr(self, '_mouseMoveMode'):
            self._mouseMoveMode = Canvas.MODE_MOUSE_NONE
        return self._mouseMoveMode

    def isCurrentlyDrawing(self, value=None):
        if value is not None:
            self._isCurrentlyDrawing = value
        if not hasattr(self, '_isCurrentlyDrawing'):
            self._isCurrentlyDrawing = False
        return self._isCurrentlyDrawing


    def getWidth(self):
        return QGLWidget.width(self)
    
    def getHeight(self):
        return QGLWidget.height(self)
    

    def initGL(self):
        # override the superclass's implementation
        # so that the GL initialization is not done in the constructor
        return


    def initializeGL(self):
        # call the superclass's actual GL initialization
        return self._initGL()


    def initViewport(self):
        # override the superclass's implementation
        # so that GL viewport initialization is not called by default
        return


    def resizeGL(self, w, h):
        # call the superclass's implementation
        # for recomputing the viewport upon resizing
        return self.resize(w, h)

    def resize(self, w, h):
        QGLWidget.resize(self, w, h)
        CanvasModule.GlCanvas.resize(self, w, h)
        

    def displayFunc(self):
        # override the superclass's implementation
        # so that GL drawing is not called by default
        return


    def paintGL(self):
        # handle re-entrancy
        if self.isCurrentlyDrawing():
            return
        self.isCurrentlyDrawing(True)
        
        # call the superclass's implementation
        # for doing the actual GL drawing
        returnVal =  self._displayFunc()

        self.isCurrentlyDrawing(False)
        return returnVal

    def OnContextMenu(self, point):

        clickedObjects = [
            x for x in 
            self.getClickableObjectsAtCanvasCoordinates(point.x(), point.y())]
        clickedObject = None
        if len(clickedObjects) is not 0:
            clickedObject = clickedObjects[0]
           
        selection = self.getSelection()[:]

        if clickedObject is None:
            pass
        elif len(selection) is 0 or not clickedObject in selection:
            selection = [clickedObject]

        import pomsets_app.gui.qt.menu as MenuModule

        # create a new popup menu
        if len(selection) is 0:
            menuClass = self.contextManager().app().getResourceValue(
                'canvas contextual menu class', 
                MenuModule.CanvasContextualMenu)
            popupMenu = menuClass(self)
            pass
        else:
            # get the contextual menu
            # from the most recent item selected
            isMultiple = any(
                [isinstance(x, GraphModule.Edge) for x in selection]) and \
                any([isinstance(x, GraphModule.Node) for x in selection])
            
            if isMultiple:
                menuClass = MenuModule.MultipleSelectionContextualMenu
                pass
            elif isinstance(selection[-1], GraphModule.Node):
                menuClass = self.contextManager().app().getResourceValue(
                    'node contextual menu class',
                    MenuModule.NodeContextualMenu)
            elif isinstance(selection[-1], GraphModule.Port):
                menuClass = self.contextManager().app().getResourceValue(
                    'port contextual menu class',
                    MenuModule.PortContextualMenu)
            elif isinstance(selection[-1], GraphModule.Edge):
                menuClass = self.contextManager().app().getResourceValue(
                    'port contextual menu class',
                    MenuModule.EdgeContextualMenu)

            popupMenu = menuClass(self)
            popupMenu.selection(selection)
            pass

        popupMenu.position(point)
        popupMenu.contextManager(self.contextManager())

        popupMenu.bindEvents()
        popupMenu.popup(self.mapToGlobal(point))

        return


    def dragMoveEvent(self, event):
        print "should process dragMoveEvent"
        return

    def dropEvent(self, event):
        print "should process dropEvent"
        return


    def mouseDoubleClickEvent(self, event):

        buttons = event.buttons()

        if buttons == Qt.LeftButton:
            eventX = event.x()
            eventY = event.y()
            clickedObjects = [
                x for x in 
                self.getClickableObjectsAtCanvasCoordinates(eventX, eventY)]
            if len(clickedObjects) is 0:
                return

            clickedNodes = [x for x in clickedObjects
                            if isinstance(x, GraphModule.Node)]
            clickedPorts = [x for x in clickedObjects
                            if isinstance(x, GraphModule.Port)]

            contextManager = self.contextManager()
            pomsetContext, pomsetReferencePath, isEditable = \
                contextManager.currentDisplayedPomsetInfo()

            if len(clickedNodes):
                # only do something if it's a nest
                clickedNode = clickedNodes[0]
                dataNode = clickedNode.nodeData
                if not dataNode.definitionToReference().isAtomic():
                    self.contextManager().mainWindow().displayPomset(
                        pomsetContext, 
                        pomsetReferencePath=pomsetReferencePath+[dataNode])
                pass
            elif len(clickedPorts):
                clickedPort = clickedPorts[0]

                dataNode = clickedPort.node.nodeData
                parameterId = clickedPort.name
                parameter = dataNode.getParameter(parameterId)

                # we don't want to allow users to edit 
                # bindings to temporal parameters
                # because bindings on them are pointless
                import pomsets.parameter as ParameterModule
                if parameter.portType() == ParameterModule.PORT_TYPE_DATA:
                    contextManager.mainWindow().OnEditParameterBinding(
                        pomsetContext, dataNode, parameter)
                pass

        return


    def mouseMoveEvent(self, event):
        buttons = event.buttons()
        keyModifiers = event.modifiers()
        mouseMoveMode= self.mouseMoveMode()

        if mouseMoveMode == Canvas.MODE_MOUSE_PAN:
            # should pan
            self.OnPanViewPort(event)
            pass
        elif mouseMoveMode == Canvas.MODE_MOUSE_ZOOM:
            # should zoom
            self.OnZoomViewPort(event)
            pass
        elif mouseMoveMode == Canvas.MODE_MOUSE_SELECTION_BOX:
            # should update the selection box
            self.OnSelectionBoundingBox(event)
            pass
        elif mouseMoveMode == Canvas.MODE_MOUSE_MOVE_OBJECTS:
            # should update the node positions
            self.OnMoveNodes(event)
            pass
        elif mouseMoveMode == Canvas.MODE_MOUSE_CREATE_EDGE:
            # should update the dynamic edge
            self.OnComputeDynamicEdge(event)
            pass

        self.updateGL()

        return


    def mouseReleaseEvent(self, event):

        mouseMoveMode = self.mouseMoveMode()
        if mouseMoveMode == Canvas.MODE_MOUSE_NONE:
            # process a drag selection from somewhere else
            pass
        elif mouseMoveMode == Canvas.MODE_MOUSE_SELECTION_BOX:
            # remove dynamic objects
            self.selection_rectangle = None
            pass
        elif mouseMoveMode == Canvas.MODE_MOUSE_CREATE_EDGE:
            # should create edge
            self.OnCreateEdge(event)
            pass

        self.mouseMoveMode(Canvas.MODE_MOUSE_NONE)
        self.updateGL()
        return


    def mousePressEvent(self, event):
        """
        Note: no need to handle contextual menu here
        """

        """
        Note: On Mac OS X, the ControlModifier value corresponds to the Command keys on the Macintosh keyboard, and the MetaModifier value corresponds to the Control keys.

Qt.NoModifier	0x00000000	No modifier key is pressed.
Qt.ShiftModifier	0x02000000	A Shift key on the keyboard is pressed.
Qt.ControlModifier	0x04000000	A Ctrl key on the keyboard is pressed.
Qt.AltModifier	0x08000000	An Alt key on the keyboard is pressed.
Qt.MetaModifier	0x10000000	A Meta key on the keyboard is pressed.
Qt.KeypadModifier	0x20000000	A keypad button is pressed.
Qt.GroupSwitchModifier	0x40000000	X11 only. A Mode_switch key on the keyboard is pressed.
        """

        eventX = event.x()
        eventY = event.y()

        buttons = event.buttons()
        keyModifiers = event.modifiers()


        self.baseMousePosition = event.pos()

        if buttons == Qt.LeftButton:

            if keyModifiers == Qt.AltModifier:
                # configure for pan
                self.initialPanMatrix = self.panMatrix
                self.mouseMoveMode(Canvas.MODE_MOUSE_PAN)
            elif keyModifiers == Qt.ShiftModifier:
                # configure for zoom
                self.initialZoomMatrix = self.zoomMatrix
                self.mouseMoveMode(Canvas.MODE_MOUSE_ZOOM)
                pass
            elif keyModifiers == Qt.ControlModifier:
                # should add to selection
                # this is the Command key on macs

                clickedObjects = [
                    x for x in 
                    self.getClickableObjectsAtCanvasCoordinates(eventX, eventY)
                    if x.isSelectable()]

                if len(clickedObjects) is 0:
                    self.initialSelection = self.getSelection()
                    self.mouseMoveMode(Canvas.MODE_MOUSE_SELECTION_BOX)
                #elif isinstance(clickedObject, GraphModule.Edge):
                #    self.mouseMoveMode(Canvas.MODE_MOUSE_CREATE_EDGE)
                else:
                    clickedObject = clickedObjects[0]
                    if clickedObject.isSelectable():
                        self.mouseMoveMode(Canvas.MODE_MOUSE_MOVE_OBJECTS)
                        self.addToSelection(clickedObjects)
                        for selected in self.getSelection():
                            if isinstance(selected, GraphModule.Edge):
                                continue
                            selected.initialX = selected.x
                            selected.initialY = selected.y
                pass
            elif keyModifiers == Qt.NoModifier:
                # should reset selection
                # and just select this object

                clickedObjects = [
                    x for x in self.getClickableObjectsAtCanvasCoordinates(eventX, eventY)]

                if len(clickedObjects) and \
                        isinstance(clickedObjects[0], GraphModule.Port):
                    self.mouseMoveMode(Canvas.MODE_MOUSE_CREATE_EDGE)
                    self.initialPort = clickedObjects[0]
                else:
                    newSelection = [x for x in clickedObjects 
                                    if x.isSelectable()]
                    if len(newSelection) is 0:
                        self.initialSelection = []
                        self.setSelection(newSelection)
                        self.mouseMoveMode(Canvas.MODE_MOUSE_SELECTION_BOX)
                    else:
                        self.mouseMoveMode(Canvas.MODE_MOUSE_MOVE_OBJECTS)
                        self.setSelection(newSelection)
                        for selected in self.getSelection():
                            if isinstance(selected, GraphModule.Edge):
                                continue
                            selected.initialX = selected.x
                            selected.initialY = selected.y

                    pass

                pass

            # attempt to select
            pass
        

        pass



    def computeMouseMoveInCanvas(self, initialPosition, finalPosition):

        initialMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % 
                                     (initialPosition.x(), initialPosition.y()))
        finalMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % 
                                   (finalPosition.x(), finalPosition.y()))

        return finalMatrix - initialMatrix


    def computeMouseMoveInWorld(self, initialPosition, finalPosition):
        baseCanvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
            (initialPosition.x(), initialPosition.y()))
        baseWorldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(
            baseCanvasMatrix)

        newCanvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
            (finalPosition.x(), finalPosition.y()))
        newWorldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(
            newCanvasMatrix)

        deltaMatrix = newWorldMatrix - baseWorldMatrix
        return deltaMatrix


    def OnZoomViewPort(self, event):
        delta = self.computeMouseMoveInCanvas(
            self.baseMousePosition, event.pos())

        delta_zoom = -0.005 * delta[3,0]

        newZoomFactor = min(max(
                self.initialZoomMatrix[0,0]+delta_zoom,0.5),10.0)

        newZoomMatrix = numpy.matrix(
            '%s 0 0 0; 0 %s 0 0; 0 0 %s 0; 0 0 0 %s' %
            (newZoomFactor, newZoomFactor, newZoomFactor, newZoomFactor))
        
        self.zoomMatrix = newZoomMatrix
        return

    def OnPanViewPort(self, event):
        deltaMatrix = self.computeMouseMoveInWorld(
            self.baseMousePosition, event.pos())

        self.panMatrix = deltaMatrix + self.initialPanMatrix
        return


    def OnMoveNodes(self, event):
        worldDeltaMatrix = self.computeMouseMoveInWorld(
            self.baseMousePosition, event.pos())

        delta = (worldDeltaMatrix[3,0], worldDeltaMatrix[3,1])

        for selected in self.getSelection():
            if isinstance(selected, GraphModule.Edge):
                continue
            x = selected.initialX + delta[0]
            y = selected.initialY + delta[1]
            selected.setPosition(x, y)
            dataNode = selected.nodeData
            guiOptions = dataNode.guiOptions()
            guiOptions['canvas position'] = (x,y)
            pass


        edgesToProcess = filter(
            lambda x: x.inputNode() in self.getSelection() or x.outputNode() in self.getSelection(),
            self.edges
        )
        map(lambda x: x.setupPrimitives(self.edgePolicy()), edgesToProcess)


        # emit this signal so that the active pomsets library
        # will know to update that the context has been modified
        self.setCurrentContextModified()
        self.emit(SIGNAL("OnNodesMoved()"))

        return


    def OnSelectionBoundingBox(self, event):

        deltaMatrix = self.computeMouseMoveInWorld(
            self.baseMousePosition, event.pos())

        baseCanvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
            (self.baseMousePosition.x(), self.baseMousePosition.y()))
        baseWorldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(
            baseCanvasMatrix)

        rect = GraphModule.Rect(
            baseWorldMatrix[3,0], baseWorldMatrix[3,1],
            deltaMatrix[3,0], deltaMatrix[3,1]
            )
        self.selection_rectangle = rect

        objectsInBoundBox = self.getObjectsIntersectingRect(
            self.selection_rectangle)

        self.setSelection(self.initialSelection + 
                          [x for x in objectsInBoundBox if x.isSelectable()])

        return


    def OnCreateEdge(self, event):

        contextManager = self.contextManager()

        clickReleaseObjects = [
            x for x in self.getClickableObjectsAtCanvasCoordinates(
            event.x(), event.y())
            if isinstance(x, GraphModule.Port)]
        clickReleaseObject = None
        if len(clickReleaseObjects) is not 0:
            clickReleaseObject = clickReleaseObjects[0]

        edge = None
        if clickReleaseObject is not None and \
           contextManager.canConnect(self.initialPort, clickReleaseObject):
            edge = contextManager.connect(self.initialPort, clickReleaseObject)

            # add the gui edge
            self.addEdge(edge)

            self.setCurrentContextModified()
            self.emit(SIGNAL("OnEdgeCreated(PyQt_PyObject)"), edge)

            pass

        # reset the dynamic edge
        # (because there won't be one anymore)
        self._dynamicEdge = None
        self._initialPort = None
        
        return


    def OnComputeDynamicEdge(self, event):
        self.computeDynamicEdge(
            GraphModule.Point(self.baseMousePosition.x(),
                              self.baseMousePosition.y()),
            GraphModule.Point(event.pos().x(), event.pos().y()))
        self._dynamicEdge.isDynamic(True)
        return


    def drawLayers(self):
        """
        This is copied over from the base class
        """
        # Draw nodes
        for layerName, layer in self.layers.iteritems():
            for child in layer.getChildren():
                if isinstance(child, GraphModule.Node):
                    self.drawNode(child)
                    continue
                if isinstance(child, GraphModule.Edge):
                    self.drawEdge(child)
                    continue
                child.draw()
        return


    def getEdgeWidth(self, baseWidth):
        # min width 1, max width 5
        width = max(1, 
                    min(int(1.0*baseWidth / self.zoomMatrix[0,0]), 
                        baseWidth))
        return width

    def drawEdge(self, edge):
        glLineWidth(self.getEdgeWidth(edge.getWidth()))

        # TODO:
        # should set the color
        visualPolicy = self.visualPolicy()
        color = edge.getColor()
        glColor4f(*color)

        if not hasattr(edge, 'children'):
            edge.setupPrimitives(self.edgePolicy())

        self.drawLine(edge.children[0])

        glLineWidth(1)
        return

    def drawDynamicEdge(self):
        # TODO:
        # consolidate the code with drawEdge
        if self._dynamicEdge is not None:
            self.drawEdge(self._dynamicEdge)
            pass
        return

    def drawNode(self, node):

        if node.shouldUpdateDisplay():
            node.updateDisplay()
            node.shouldUpdateDisplay(False)

        glPushMatrix()
        glTranslate(node.x, node.y, 0)

        for nodePrimitive in node.partPrimitives():
            if not nodePrimitive.visible():
                continue

            if isinstance(nodePrimitive, GraphModule.Rect):
                self.drawRect(nodePrimitive)
                continue

            if isinstance(nodePrimitive, GraphModule.Text):
                self.drawText(nodePrimitive)
                continue

            if isinstance(nodePrimitive, GraphModule.Port):
                self.drawPort(nodePrimitive)
                continue

            nodePrimitive.draw()
            
        glPopMatrix()
        return


    def drawPort(self, port):
        glPushMatrix()
        glTranslate(port.x, port.y, 0)

        for child in port.children:
            if isinstance(child, GraphModule.Rect):
                glColor4f(*child.borderColor)
                self.drawRect(child, filled=child.filled())
                continue
            if isinstance(child, GraphModule.Circle):
                glColor4f(*child.borderColor)
                self.drawCircle(child)
                continue

            child.draw()

        glPopMatrix()
        return
    


    def drawText(self, primitive):
        color = primitive.color
        color[3] = color[3] * primitive.alpha
        glColor4f(*color)

        fontDb = self.contextManager().app()._fontDb
        fontSize = int(18.0 / self.zoomMatrix[0,0])

        # TODO:
        # this should use the visual policy instead
        font = fontDb.font('FreeUniversal', 'Normal', 18)

        zoom = self.zoomMatrix[0, 0]

        fontMetrics = QtGui.QFontMetrics(font)
        rect = fontMetrics.boundingRect(primitive.text)
        
        offsetX = primitive.position[0] + (primitive.size[0]-rect.width())/2
        offsetY = primitive.position[1] + (primitive.size[1]-rect.height())/2

        font = fontDb.font('FreeUniversal', 'Normal', fontSize)
        self.renderText(offsetX, offsetY, 0,
                        primitive.text,
                        font)
        return


    def drawCircle(self, circle, *args, **kwds):

        # we want at least 8 wedges
        # and for now, a max 
        smoothness = max(8, circle.size[0])
        smoothness = min(smoothness, 360)
        
        x = circle.size[0]/2.0
        y = circle.size[1]/2.0
        radius = circle.radius
            
        glBegin(GL_LINE_LOOP)
        for i in range(0, smoothness):    
            angle = i * math.pi * 2.0 / smoothness
            glVertex2f(x + radius * math.cos(angle), 
                       y + radius * math.sin(angle))
        glEnd() 

        return



    def drawRect(self, rect, *args, **kwds):
        if hasattr(rect, 'use_texture') and rect.use_texture:
            textureFile = rect._texture_image_file
            if not textureFile in self.textureMap:
                image = QtGui.QImage(textureFile)
                textureId = self.bindTexture(image)
                self.textureMap[textureFile] = (textureId, image)
            else:
                textureId, image = self.textureMap[textureFile]
                pass

            # NOTE:
            # this is a hack to get the PS image drawing
            # TODO:
            # find a better solution

            # cache the value so that it does not get changed
            shouldTranslate = False
            if hasattr(rect, 'offsetPosition'):
                shouldTranslate=True

            if shouldTranslate:
                glPushMatrix()
                glTranslate(*rect.offsetPosition)

            textureRect = QRectF(0.0, 0.0, rect.width, rect.height)
            self.drawTexture(textureRect, textureId)

            if shouldTranslate:
                glPopMatrix()

        elif isinstance(rect, GraphModule.Rect):
            CanvasModule.GlCanvas.drawRect(self, rect, *args, **kwds)
            pass
        else:
            rect.draw()
        return


    def OnRefresh(self):
        pomsetContext, pomsetReferencePath, isEditable = \
            self.contextManager().currentDisplayedPomsetInfo()

        self.displayPomset(pomsetReferencePath)

        self.updateGL()
        return


    def OnNodesDuplicated(self, nodeMap):
        for nodeCopy, originalNode in nodeMap.iteritems():
            # TODO:
            # this should be relative to the gui node
            # rather than hardcoded to 0, 0
            guiOptions = originalNode.guiOptions()
            if 'canvas position' in guiOptions:
                copyGuiOptions = nodeCopy.guiOptions()
                x, y = guiOptions['canvas position']
                x = x + 25
                y = y - 25
                copyGuiOptions['canvas position'] = (x,y)
            pass

        # rebuild the gui objects
        pomsetContext, pomsetReferencePath, isEditable = \
            self.contextManager().currentDisplayedPomsetInfo()
        self.displayPomset(pomsetReferencePath)

        # update the selection to the duplicated objects
        guiNodes = map(self.getGuiObjectForDataNode, nodeMap.keys())
        self.setSelection(guiNodes)

        # redraw
        self.updateGL()
        return


    def OnNodeCreated(self, node, pos):
        uiNode = self.addNode(node)
        uiNodeSize = uiNode.getBounds()

        canvasMatrix = numpy.matrix(
            '1 0 0 0; 0 1 0 0; 0 0 1 0; %s %s 0 1' % 
            (pos.x() - uiNodeSize[0]/2.0,
             pos.y() + uiNodeSize[1]/2.0)
        )
        worldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(
            canvasMatrix)

        position = (worldMatrix[3,0], worldMatrix[3,1])
        guiOptions = node.guiOptions()
        guiOptions['canvas position'] = position
        uiNode.setPosition(*position)

        self.updateGL()
        return




    def drawMetaInfo(self):
        
        CanvasModule.GlCanvas.drawMetaInfo(self)

        # draw breadcrumbs
        x = 30
        y = self.height - 30
        glColor3f(1.0, 1.0, 1.0)
        fontSize = 12
        fontDb = self.contextManager().app()._fontDb
        font = fontDb.font('Arial', 'Normal', fontSize)
        self.renderText(x, y, 0,
                        ">>> " + self.breadCrumbs(),
                        font)
        return


    def getScaledPosition(self, rawPosition):
        positionString = ' '.join([str(x) for x in rawPosition+[0]])
        rawPositionMatrix = numpy.matrix(positionString)
        scaledPositionMatrix = rawPositionMatrix / self.zoomMatrix
        return [scaledPositionMatrix[0,0],
                scaledPositionMatrix[1,1],
                scaledPositionMatrix[2,2]]
        

    def event(self, event):
        if event.type() == QEvent.ToolTip:
            position = event.pos()

            clickedObjects = [
                x for x in 
                self.getClickableObjectsAtCanvasCoordinates(
                    position.x(), position.y())]

            clickedObject = None
            if len(clickedObjects) is not 0:
                clickedObject = clickedObjects[0]
                if not isinstance(clickedObject, GraphModule.Port):
                    clickedObject = None
                pass

            if clickedObject is not None:
                QtGui.QToolTip.showText(event.globalPos(),
                                        clickedObject.name)
            else:
                QtGui.QToolTip.hideText()
                event.ignore()
            return True
        return QGLWidget.event(self, event)

    # END class Canvas
    pass






import logging
import os

from PyQt4.QtCore import *
from PyQt4 import QtGui

import pypatterns.command as CommandModule
import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule
import pypatterns.relational.commands as RelationalCommandModule

import pomsets.parameter as ParameterModule

import pomsets_app.utils as AppUtilsModule
import pomsets_app.gui.qt as QtModule
import pomsets_app.controller.automaton as AutomatonModule
import pomsets_app.controller.commands as CloudCommandModule


class Controller(QtModule.Controller, QtGui.QDialog):

    INDEX_INSTANCEID = 3

    def __init__(self, *args, **kwds):
        QtGui.QDialog.__init__(self, *args, **kwds)
        QtModule.Controller.__init__(self)
        return

    def populate(self):

        widget = self.widget()
        contextManager = self.contextManager()

        self.updateGuiTable()
        return


    @pyqtSlot(QtGui.QAbstractButton)
    def on_buttonBox_clicked(self, button):
        widget = self.widget()
        buttonRole = widget.buttonBox.buttonRole(button)
        if buttonRole == QtGui.QDialogButtonBox.ApplyRole:
            self.saveValues()
            pass
        if buttonRole == QtGui.QDialogButtonBox.AcceptRole:
            self.saveValues()
            self.accept()
            pass
        if buttonRole == QtGui.QDialogButtonBox.RejectRole:
            self.reject()
            pass
        return

    @pyqtSlot()
    def on_buttonCredentials_clicked(self):
        self.contextManager().mainWindow().OnEuca2oolsConfiguration()
        return

    @pyqtSlot()
    def on_buttonAssign_clicked(self):
        self.OnAssignSelected()
        return

    @pyqtSlot()
    def on_buttonUnassign_clicked(self):
        self.OnUnassignSelected()
        return

    @pyqtSlot()
    def on_buttonRefresh_clicked(self):
        self.refreshDataModel()
        self.updateGuiTable()
        return



    def getSelectedVmInstances(self):
        # get the thread to be dismissed
        # get the vm to be terminated
        # (only if there are no other threads)

        widget = self.widget()
        tableInstances = widget.tableInstances
        selectedItems = tableInstances.selectedItems()

        selectedRows = list(set([x.row() for x in selectedItems]))
        selectedInstances = []
        for selectedRow in selectedRows:
            item = tableInstances.item(selectedRow, Controller.INDEX_INSTANCEID)
            selectedInstances.append(str(item.data(Qt.DisplayRole).toString()))
        return selectedInstances


    def OnAssignSelected(self):

        contextManager = self.contextManager()
        automaton = contextManager.automaton()

        selectedVmInstances = self.getSelectedVmInstances()

        command = CommandModule.CompositeCommand()
        for vmInstance in selectedVmInstances:

            executeEnvironments = \
                automaton.getExecuteEnvironmentsFromVmInstance(vmInstance)
            if len(executeEnvironments) is not 0:
                continue
            serviceAPI = automaton.getServiceApiFromVmInstance(vmInstance)

            assignEnvCommand = \
                CloudCommandModule.AssignExecuteEnvironmentForVmCommand(
                automaton, serviceAPI, vmInstance)

            command.addCommand(assignEnvCommand)
            pass


        executeEnvironments = []
        if len(command._subcommands):

            # execute the command
            success = automaton.executeCommand(command)

            # if not successful
            if not success:
                logging.error('failed creating execute environments for vmInstances %s' % selectedVmInstances)
                return
            pass
        else:
            logging.error('no commands were generated for assigning execute environments')

        executeEnvironments = {}
        for vmInstance in selectedVmInstances:

            executeEnvironments[vmInstance] = \
                automaton.getExecuteEnvironmentsFromVmInstance(vmInstance)
            pass


        if len(executeEnvironments) is 0:
            logging.error('cannot assign threads as no execute environments are selected')
            return


        command = CommandModule.CompositeCommand()

        for vmInstance, executeEnvironmentList in executeEnvironments.iteritems():

            if len(executeEnvironmentList) is not 1:
                raise NotImplementedError('need to handle multiple execute environments for the same vm instance')
            executeEnvironment = executeEnvironmentList[0]

            threads = automaton.getThreadsFromExecuteEnvironment(executeEnvironment)
            if len(threads) is not 0:
                # TODO:
                # will need a way to create multiple threads
                # for the same execute environment
                continue

            # TODO:
            # add ability to parameterize the threadPoolId
            threadPoolId = AutomatonModule.ID_EXECUTE_EUCA2OOLS

            assignWorkerCommand = \
                CloudCommandModule.AssignThreadForExecuteEnvironmentCommand(
                automaton, threadPoolId, executeEnvironment
                )

            command.addCommand(assignWorkerCommand)

            rowFilter = RelationalModule.ColumnValueFilter(
                'Instance ID',
                FilterModule.EquivalenceFilter(vmInstance))

            # TODO:
            # this should be a command builder
            # but we don't have a command builder right now
            # that sets a column using a row filter
            command.addCommand(
                RelationalCommandModule.SetColumnValueUsingRowFilterCommand(
                    rowFilter,
                    automaton.virtualMachineTable(),
                    'Enabled',
                    str(True)
                    )
                )
            

            pass


        success = automaton.executeCommand(command)
        if success:
            self.updateGuiTable()


        # update the execute menu items
        # contextManager.app().GetTopWindow().updateExecuteMenus()


        return

    def OnUnassignSelected(self):

        contextManager = self.contextManager()
        automaton = contextManager.automaton()
        dataModel = automaton.virtualMachineTable()

        selectedVmInstances = self.getSelectedVmInstances()

        vmInstances = []

        for vmInstance in selectedVmInstances:
            filter = FilterModule.constructAndFilter()
            filter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'Instance ID',
                    FilterModule.EquivalenceFilter(vmInstance)
                    )
                )
            filter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'Enabled',
                    FilterModule.EquivalenceFilter('True'))
                )
            rows = [x for x in dataModel.retrieve(filter)]
            if len(rows):
                vmInstances.append(vmInstance)
            pass


        if len(vmInstances) is 0:
            logging.debug("nothing to do, as no rows are selected")
            return

        command = CommandModule.CompositeCommand()

        # TODO:
        # add ability to parameterize the threadPoolId
        threadPoolId = AutomatonModule.ID_EXECUTE_EUCA2OOLS

        for vmInstance in vmInstances:
            # create the command to dismiss the worker
            # but do it only if there's a worker assigned to the vm
            # which there should be
            dismissWorkerCommand = CloudCommandModule.UnassignWorkerFromVmCommand(
                automaton, threadPoolId, vmInstance
                )
            command.addCommand(dismissWorkerCommand)


            rowFilter = RelationalModule.ColumnValueFilter(
                'Instance ID',
                FilterModule.EquivalenceFilter(vmInstance))

            # TODO:
            # this should be a command builder
            # but we don't have a command builder right now
            # that sets a column using a row filter
            command.addCommand(
                RelationalCommandModule.SetColumnValueUsingRowFilterCommand(
                    rowFilter,
                    automaton.virtualMachineTable(),
                    'Enabled',
                    str(False)
                    )
                )

            pass

        success = automaton.executeCommand(command)
        if success:
            self.updateGuiTable()

        return


    def refreshDataModel(self):


        contextManager = self.contextManager()
        automaton = contextManager.automaton()

        dataModel = automaton.virtualMachineTable()
        rows = dataModel.clear()

        vmEnvs = {}
        envThreads = {}
        for row in rows:
            vmInstance = row.getColumn('Instance ID')
            executeEnvironments = \
                automaton.getExecuteEnvironmentsFromVmInstance(vmInstance)
            if len(executeEnvironments) is 0:
                continue

            vmEnvs[vmInstance] = executeEnvironments[0]
            threads = automaton.getThreadsFromExecuteEnvironment(
                executeEnvironments[0])
            if len(threads) is 0:
                continue

            envThreads[executeEnvironments[0]] = threads[0]

            # TODO:
            # need to handle assigned tasks

            pass

        reservations = []
        try:
            controllerValues = \
                automaton.getCloudControllerCredentialsForAPI('euca2ools')
        except KeyError, e:
            logging.warn("should query user for credentials")
            pass


        try:

            # 'Service', 'Reservation ID', 'Instance ID', 'Image ID', 'IP', 
            # 'Enabled', 'Thread', 'Assigned task'

            from euca2ools import Euca2ool
            euca = Euca2ool()

            euca.ec2_url = controllerValues['url']
            euca.ec2_user_access_key = controllerValues['access key']
            euca.ec2_user_secret_key = controllerValues['secret key']

            euca_conn = euca.make_connection()

            reservations = euca_conn.get_all_instances([])
        except Exception, e:
            logging.error("error on making connecting to Eucalyptus >> %s" % e)
            pass

        if len(reservations) is 0:
            logging.debug("no active reservations")

        rows = []
        for reservation in reservations:
            for instance in reservation.instances:
                # reservation.id
                # instance.id
                # instance.image_id
                # instance.public_dns_name
                # instance.state (this is so we dont show terminated ones)
                # instance.key_name
                row = dataModel.addRow()
                row.setColumn('Service', 'Eucalyptus')
                row.setColumn('API', 'euca2ools')
                row.setColumn('Reservation ID', reservation.id)
                row.setColumn('Instance ID', instance.id)
                row.setColumn('Image ID', instance.image_id)
                row.setColumn('IP', instance.public_dns_name)
                row.setColumn('Key name', instance.key_name)


                # TODO:
                # this should be a join operation


                vmEnv = vmEnvs.get(instance.id, None)
                envThread = None
                if vmEnv:
                    if envThread:
                        envThread = envThreads.get(vmEnv, None)
                    row.setColumn('Enabled', 'True')
                else:
                    row.setColumn('Enabled', 'False')

                rows.append(row)
                pass
            pass
        return



    def updateGuiTable(self):

        widget = self.widget()
        contextManager = self.contextManager()
        automaton = contextManager.automaton()

        tableInstances = widget.tableInstances
        dataModel = automaton.virtualMachineTable()

        # first clear the table
        while tableInstances.rowCount():
            tableInstances.removeRow(0)


        tableInstances.setRowCount(dataModel.rowCount())
        
        for rowIndex, row in enumerate(dataModel.rows()):

            item = QtGui.QTableWidgetItem('Service')
            item.setData(Qt.DisplayRole, row.getColumn('Service'))
            tableInstances.setItem(rowIndex, 0, item)
            
            item = QtGui.QTableWidgetItem('API')
            item.setData(Qt.DisplayRole, row.getColumn('API'))
            tableInstances.setItem(rowIndex, 1, item)

            item = QtGui.QTableWidgetItem('Reservation ID')
            item.setData(Qt.DisplayRole, row.getColumn('Reservation ID'))
            tableInstances.setItem(rowIndex, 2, item)

            item = QtGui.QTableWidgetItem('Instance ID')
            item.setData(Qt.DisplayRole, row.getColumn('Instance ID'))
            tableInstances.setItem(rowIndex, Controller.INDEX_INSTANCEID, item)

            item = QtGui.QTableWidgetItem('Image ID')
            item.setData(Qt.DisplayRole, row.getColumn('Image ID'))
            tableInstances.setItem(rowIndex, 4, item)

            item = QtGui.QTableWidgetItem('IP address')
            item.setData(Qt.DisplayRole, row.getColumn('IP'))
            tableInstances.setItem(rowIndex, 5, item)

            item = QtGui.QTableWidgetItem('Key name')
            item.setData(Qt.DisplayRole, row.getColumn('Key name'))
            tableInstances.setItem(rowIndex, 6, item)

            item = QtGui.QTableWidgetItem('Assigned')
            item.setData(Qt.DisplayRole, row.getColumn('Enabled'))
            tableInstances.setItem(rowIndex, 7, item)
            pass

        return



    # END class Controller
    pass



import wx
import wx.grid

import pypatterns.relational as RelationalModule
import pomsets.resource as ResourceModule

class Table(wx.grid.PyGridTableBase, ResourceModule.Struct):
    
    ATTRIBUTES = ['dataModel', 'columns', 'columnIndices']
    
    def __init__(self, dataModel, columns=None):
        wx.grid.PyGridTableBase.__init__(self)
        
        self.dataModel(dataModel)
        if columns is None:
            columns = dataModel.columns().keys()

        # at this point, columns is a list of column names
        self.columns(columns)
        self.columnIndices(dict(enumerate(columns)))

        return
    
    def GetNumberRows(self):
        """Return the number of rows in the grid"""
        return len(self.dataModel().rows())

    def GetNumberCols(self):
        """Return the number of columns in the grid"""
        return len(self.columns())

    def IsEmptyCell(self, rowIndex, col):
        """Return True if the cell is empty"""
        try:
            dataRow = self.dataModel().rows()[rowIndex]
        except IndexError:
            return True
        return False

    def getIndexForColumn(self, columnName):
        return self.columns().index(columnName)
    
    def getObjectAt(self, rowIndex, colIndex):
        try:
            dataRow = self.dataModel().rows()[rowIndex]
            columnName = self.columnIndices()[colIndex]
            obj = dataRow.getColumn(columnName)
            return obj
        except IndexError:
            # need to catch this because of a bug with PyGridTableBase
            pass
        return ''
    
    def GetTypeName(self, row, col):
        """Return the name of the data type of the value in the cell"""
        obj = self.getObjectAt(row, col)
        
        if type(obj) is str:
            return None
        
        return obj.__class__.__name__

    def GetValueAsCustom(self, row, col, typeName):
        
        if self.GetTypeName(row, col) == typeName:
            obj = self.getObjectAt(row, col)
            return obj
        
        return self.GetValue(row, col)
    
    
    def GetValue(self, row, col):
        """Return the value of a cell"""
        
        obj = self.getObjectAt(row, col)
        
        # TODO:
        # there needs to be something that converts
        # the object in the data model into the value to be displayed
        
        return obj
    

    def SetValue(self, row, col, value):
        """Set the value of a cell"""
        dataRow = self.dataModel().rows()[row]
        columnName = self.columnIndices()[col]
        
        # TODO:
        # there needs to be something that converts 
        # the value displayed into the object in the data model
        
        dataRow.setColumn(columnName, value)
        pass


    def GetColLabelValue(self, col):
        return self.columnIndices()[col]
    
    
    # END class Table
    pass




import functools
import logging
import os
import platform
import sys
import uuid

import numpy

from PyQt4 import QtGui
from PyQt4.QtCore import *
from PyQt4.QtOpenGL import *


import pypatterns.relational as RelationalModule
import pypatterns.filter as FilterModule

import pomsets.context as ContextModule
import pomsets.library as LibraryModule

import pomsets_app.utils as AppUtilsModule
import pomsets_app.controller.automaton as AutomatonModule
import pomsets_app.gui.frame as FrameModule
import pomsets_app.gui.utils as GuiUtilsModule
import pomsets_app.gui.qt.canvas as CanvasModule
import pomsets_app.gui.qt.menu as MenuModule
import pomsets_app.gui.qt.pomset_tree.widget as PomsetTreeModule


class Frame(QtGui.QMainWindow, FrameModule.Frame):

    def __init__(self):
        QtGui.QMainWindow.__init__(self)
        self.setWindowTitle('pomsets')
        self.setMinimumSize(800, 600)
        self._widgets = {}
        return

    def putWidget(self, name, widget):
        self._widgets[name] = widget
        return

    def getWidget(self, name):
        return self._widgets[name]


    def initializeMenuBar(self):
        menuBar = self.menuBar()

        pomsetMenu = menuBar.addMenu('Pomsets')
        newPomsetAction = pomsetMenu.addAction('New')
        loadPomsetAction = pomsetMenu.addAction('Load')
        savePomsetAction = pomsetMenu.addAction('Save')
        savePomsetAsAction = pomsetMenu.addAction('Save as')
        closePomsetAction = pomsetMenu.addAction('Close')
        self.connect(newPomsetAction,
                     SIGNAL('triggered()'), self.OnCreateNewTopLevelPomset)
        self.connect(loadPomsetAction, 
                     SIGNAL('triggered()'), self.OnLoadTopLevelPomset)
        self.connect(savePomsetAction, 
                     SIGNAL('triggered()'), self.OnSaveCurrentActivePomset)
        self.connect(savePomsetAsAction, 
                     SIGNAL('triggered()'), self.OnSavePomset)
        self.connect(closePomsetAction, 
                     SIGNAL('triggered()'), self.OnCloseActivePomset)


        # we only do this if not a Mac
        # because on a mac, the exit and about menus
        # get added to the bar at the top of the windw
        if not platform.system() in ['Darwin']:
            pomsetMenu.addSeparator()
        exitApplicationAction = pomsetMenu.addAction('Exit')
        self.connect(exitApplicationAction, 
                     SIGNAL('triggered()'), self.OnExitApplication)

        
        # we only do this if not a Mac
        # because on a mac, the exit and about menus
        # get added to the bar at the top of the windw
        if not platform.system() in ['Darwin']:
            helpMenu = menuBar.addMenu('Help')
            aboutApplicationAction = helpMenu.addAction('About')
        else:
            aboutApplicationAction = pomsetMenu.addAction('About')
        self.connect(aboutApplicationAction, 
                     SIGNAL('triggered()'), self.OnAboutApplication)


        executionMenu = menuBar.addMenu('Execute')
        previewExecutionAction = executionMenu.addAction('Preview')
        executionMenu.addSeparator()
        localExecutionAction = executionMenu.addAction('Local')
        eucaExecutionAction = executionMenu.addAction('Cloud')

        self.connect(previewExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecutePreview)
        self.connect(localExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteLocal)
        self.connect(eucaExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteEucalyptus)


        configurationMenu = menuBar.addMenu('Configuration')
        cloudpoolAction = configurationMenu.addAction('Cloud pool')
        configurationMenu.addSeparator()
        euca2oolsAction = configurationMenu.addAction('Eucalyptus')
        hadoopAction = configurationMenu.addAction('Hadoop')

        self.connect(cloudpoolAction, 
                     SIGNAL('triggered()'), self.OnCloudPoolConfiguration)
        self.connect(euca2oolsAction, 
                     SIGNAL('triggered()'), self.OnEuca2oolsConfiguration)
        self.connect(hadoopAction, 
                     SIGNAL('triggered()'), self.OnHadoopConfiguration)

        return


    def initializeWidgets(self):

        self.initializeMenuBar()

        mainSplitterWidget = QtGui.QSplitter(self)

        librarySplitterWidget = QtGui.QSplitter(mainSplitterWidget)
        librarySplitterWidget.setOrientation(Qt.Vertical)

        activeLibraryTreeWidget = PomsetTreeModule.PomsetTreeWidget(librarySplitterWidget)
        activeLibraryTreeWidget.setHeaderLabels(['Active pomsets'])
        activeLibraryTreeWidget.setHeaderHidden(False)
        librarySplitterWidget.addWidget(activeLibraryTreeWidget)
        self.putWidget('active library tree', activeLibraryTreeWidget)

        toolbox = QtGui.QToolBox(librarySplitterWidget)
        toolbox.addItem(PomsetTreeModule.PomsetTreeWidget(toolbox), 'Loaded definitions')
        toolbox.addItem(PomsetTreeModule.PomsetTreeWidget(toolbox), 'Library definitions')
        toolbox.setCurrentIndex(1)
        librarySplitterWidget.addWidget(toolbox)

        canvas = CanvasModule.Canvas(mainSplitterWidget, 800, 600)
        canvas.contextManager(self.app().contextManager())
        canvas.resetDrawables()
        self.putWidget('graph canvas', canvas)
        self.canvas = canvas
        mainSplitterWidget.addWidget(canvas)

        self.setCentralWidget(mainSplitterWidget)

        # signal from popup menu to create node
        QObject.connect(
            toolbox.widget(0),
            SIGNAL("OnCreateNode(PyQt_PyObject,PyQt_PyObject,PyQt_PyObject,QPoint)"),
            self.OnCreateNode)
        QObject.connect(
            toolbox.widget(1), 
            SIGNAL("OnCreateNode(PyQt_PyObject,PyQt_PyObject,PyQt_PyObject,QPoint)"),
            self.OnCreateNode)
        
        # notify the widgets that a node got created
        QObject.connect(self, SIGNAL("OnNodeCreated(PyQt_PyObject,QPoint)"),
                        canvas.OnNodeCreated)
        QObject.connect(self, SIGNAL("OnNodeCreated(PyQt_PyObject,QPoint)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widgets that an edge got created
        QObject.connect(canvas, SIGNAL("OnEdgeCreated(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widgets that a node got moved
        QObject.connect(canvas, SIGNAL("OnNodesMoved()"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widgets that a node got duplicated
        QObject.connect(self, SIGNAL("OnNodesDuplicated(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)
        QObject.connect(self, SIGNAL("OnNodesDuplicated(PyQt_PyObject)"),
                        canvas.OnNodesDuplicated)

        # notify the widgets that a node got deleted
        QObject.connect(self, SIGNAL("OnNodesDeleted()"),
                        activeLibraryTreeWidget.OnRefresh)
        QObject.connect(self, SIGNAL("OnNodesDeleted()"),
                        canvas.OnRefresh)
        
        # notify the widget that a pomset got moved across libraries
        QObject.connect(self, SIGNAL("OnPomsetMoved()"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetMoved()"),
                        toolbox.widget(1).OnRefresh)


        # notify the widget that a pomset got created
        QObject.connect(self, SIGNAL("OnPomsetCreated()"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetCreated()"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetCreated()"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset got loaded
        QObject.connect(self, SIGNAL("OnPomsetLoaded(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetLoaded(PyQt_PyObject)"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetLoaded(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset got saved
        QObject.connect(self, SIGNAL("OnPomsetSaved(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetSaved(PyQt_PyObject)"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetSaved(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset was closed
        QObject.connect(self, SIGNAL("OnPomsetClosed(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetClosed(PyQt_PyObject)"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetClosed(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset got modified
        QObject.connect(self, SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        canvas.OnRefresh)

        # TODO:
        # notify that a pomset was submitted for execution

        # TODO:
        # notify that a pomset errored in execution

        # TODO:
        # notify a task execution status changed
        QObject.connect(self, SIGNAL("OnTaskExecutionStatusChanged()"),
                        canvas.updateGL)

        return


    def closeEvent(self, event):
        # call our own exit application function
        self.OnExitApplication()
        event.ignore()
        return


    def setWidgetDataSources(self):

        contextManager = self.app().contextManager()

        mainSplitterWidget = self.centralWidget()

        librarySplitterWidget = mainSplitterWidget.widget(0)
        toolboxWidget = librarySplitterWidget.widget(1)
        
        persistentLibraryTreeWidget = toolboxWidget.widget(1)
        
        persistentLibraryTreeWidget.pomsetLibrary(
            contextManager.persistentLibrary())
        persistentLibraryTreeWidget.emptyDataSourceString('None loaded')
        persistentLibraryTreeWidget.contextManager(contextManager)


        notBootstrapFilter = FilterModule.constructNotFilter()
        notBootstrapFilter.addFilter(
            LibraryModule.getBootstrapLoaderPomsetsFilter())
        persistentLibraryTreeWidget.pomsetFilter(notBootstrapFilter)
        persistentLibraryTreeWidget.shouldEnableDragToCanvas(True)
        persistentLibraryTreeWidget.contextualMenuClass(
            MenuModule.LibraryDefinitionContextualMenu)

        transientLibraryTreeWidget = toolboxWidget.widget(0)
        transientLibraryTreeWidget.contextManager(contextManager)
        transientLibraryTreeWidget.pomsetLibrary(
            contextManager.transientLibrary())
        transientLibraryTreeWidget.emptyDataSourceString('None loaded')
        transientLibraryTreeWidget.shouldIndicateModified(True)
        transientLibraryTreeWidget.shouldEnableDragToCanvas(True)
        transientLibraryTreeWidget.contextualMenuClass(
            MenuModule.LoadedDefinitionContextualMenu)

        activeLibraryTreeWidget = librarySplitterWidget.widget(0)
        activeLibraryTreeWidget.contextManager(contextManager)
        activeLibraryTreeWidget.pomsetLibrary(
            contextManager.activeLibrary())
        activeLibraryTreeWidget.emptyDataSourceString('None active')
        activeLibraryTreeWidget.shouldShowInternals(True)
        activeLibraryTreeWidget.shouldExpand(True)
        activeLibraryTreeWidget.shouldIndicateModified(True)
        activeLibraryTreeWidget.contextualMenuClass(
            MenuModule.ActivePomsetContextualMenu)
        activeLibraryTreeWidget.mouseDoubleClickCallback(
            functools.partial(activeLibraryTreeWidget.displayPomset,
                              frame=self))

        return

    def populateWidgets(self):

        mainSplitterWidget = self.centralWidget()

        librarySplitterWidget = mainSplitterWidget.widget(0)
        toolboxWidget = librarySplitterWidget.widget(1)

        widgets = [librarySplitterWidget.widget(0)] + \
            map(toolboxWidget.widget, range(toolboxWidget.count()))
            
        for widget in widgets:
            widget.populate()

        return


    def OnSaveCurrentActivePomset(self):
        contextManager = self.app().contextManager()
        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        fileName = pomsetContext.url()
        self.OnSavePomset(pomsetContext, fileName)
        return


    def OnSavePomset(self, pomsetContext=None, fileName=None):
        contextManager = self.app().contextManager()

        if pomsetContext is None:
            pomsetContext, pomsetReferencePath, isEditable = \
                contextManager.currentDisplayedPomsetInfo()

        if fileName is None:
            import user
            fileName = QtGui.QFileDialog.getSaveFileName(
                self,
                QString("Save pomset"),
                QString(user.home),
                QString("Pomset workflows (*.pomset)"))

        if fileName and len(fileName):
            ContextModule.savePomsetAs(pomsetContext, str(fileName))
            self.emit(SIGNAL("OnPomsetSaved(PyQt_PyObject)"), pomsetContext)
            return True

        return False


    def OnLoadTopLevelPomset(self):

        contextManager = self.app().contextManager()
        pomsetContext = self.OnLoadPomsetIntoLibrary(
            contextManager.activeLibrary())
        if pomsetContext:
            contextManager.activePomsetContext(pomsetContext)
            self.displayPomset(pomsetContext)
            # self.canvas.OnRefresh()
            self.canvas.updateGL()
        return


    def OnLoadPomsetIntoLibrary(self, library, filter=None):
        import user
        fileName = QtGui.QFileDialog.getOpenFileName(
            self,
            QString("Load pomset"), 
            QString(user.home),
            QString("Pomset workflows (*.pomset)"))

        # TODO:
        # remove this hack that currently
        # allows us to prevent loading nest definitions
        if filter is None:
            filter = FilterModule.TRUE_FILTER

        pomsetContext = None
        if fileName and len(fileName):
            pomsetContext = ContextModule.loadPomset(path=fileName)
            if filter.matches(pomsetContext):
                library.addPomsetContext(pomsetContext)
                self.emit(SIGNAL("OnPomsetLoaded(PyQt_PyObject)"), 
                          pomsetContext)
            else:
                # TODO:
                # need to display an error message
                # to the user
                pass
            pass

        return pomsetContext

    def OnRenameTopLevelPomset(self, newName):

        contextManager = self.app().contextManager()
        pomsetContext = contextManager.activePomsetContext()
        pomset = pomsetContext.pomset()

        pomset.name(newName)

        self.displayPomset(pomsetContext)
        self.canvas.OnRefresh()
        self.emit(SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return

    def OnCreateNewTopLevelPomset(self):

        contextManager = self.app().contextManager()
        pomsetContext = contextManager.createNewPomset()
        contextManager.addActivePomsetContext(pomsetContext)
        self.displayPomset(pomsetContext)
        self.canvas.OnRefresh()
        self.emit(SIGNAL("OnPomsetCreated()"))
        return


    def OnCreateNewAtomicPomset(self):

        name, status = QtGui.QInputDialog.getText(
            self,
            QString("Create definition"),
            QString("Name:"))

        # create the node only if the user provided a name
        # and did not cancel the request
        if status:
            name = str(name)
            context = GuiUtilsModule.createAtomicDefinition(
                self.app().contextManager(),
                name=name)

            self.emit(SIGNAL("OnPomsetCreated()"))

        return

    def OnCreateNewNestPomset(self):
        
        name, status = QtGui.QInputDialog.getText(
            self,
            QString("Create nest"),
            QString("Name:"))

        # create the node only if the user provided a name
        # and did not cancel the request
        if status:
            name = str(name)
            #context = GuiUtilsModule.createNestDefinition(
            #    self.app().contextManager(),
            #    name=name)
            # self.emit(SIGNAL("OnPomsetCreated()"))
            print "should create new nest"

        return


    def OnCloseActivePomset(self, pomsetContext=None, 
                            shouldQueryIfModified=True):

        contextManager = self.app().contextManager()

        if pomsetContext is None:
            pomsetContext, pomsetReferencePath, isEditable = \
                contextManager.currentDisplayedPomsetInfo()
        
        pomsetReference = pomsetContext.reference()

        # query the user to save the modification
        if pomsetContext.isModified() and shouldQueryIfModified:
            pomsetName = pomsetReference.name()

            msgBox = QtGui.QMessageBox(self)
            msgBox.setText("pomset \"%s\" has been modified." % pomsetName)
            msgBox.setInformativeText("Discard changes, save, or cancel close?")
            msgBox.setStandardButtons(
                QtGui.QMessageBox.Discard | 
                QtGui.QMessageBox.Save | 
                QtGui.QMessageBox.Cancel)
            msgBox.setDefaultButton(QtGui.QMessageBox.Cancel)
            ret = msgBox.exec_()

            if ret == QtGui.QMessageBox.Cancel:
                return
            if ret == QtGui.QMessageBox.Save:
                pomsetSaved = self.OnSavePomset(pomsetContext=pomsetContext, 
                                                fileName=pomsetContext.url())
                if not pomsetSaved:
                    # the user cancled the file dialog
                    # so we assume that they want to cancel the save
                    return
                pass
            if ret == QtGui.QMessageBox.Discard:
                # discard does nothing, 
                # we continue with the rest of this function
                pass

            pass

        contextManager.closePomset(pomsetContext)

        activeContext = contextManager.activePomsetContext()
        self.displayPomset(activeContext)
        self.canvas.OnRefresh()

        self.emit(SIGNAL("OnPomsetClosed(PyQt_PyObject)"), pomsetContext)
        return

    def OnDuplicateSelection(self, selection):

        import pomsets_app.gui.graph as GraphModule
        nodes = [x for x in selection if isinstance(x, GraphModule.Node)]
        edges = [x for x in selection if isinstance(x, GraphModule.Edge)]

        # first duplicate the nodes
        nodeMap = self.OnDuplicateNodes(nodes)

        # now connect the edges
        self.OnDuplicateEdges(edges, nodeMap)
        return

    def OnDuplicateEdges(self, edges, nodeMap):
        
        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        for guiEdge in edges:
            sourceNode = nodeMap[guiEdge.inputNode().nodeData]
            sourceParameterId = guiEdge.inputPort().name
            targetNode = nodeMap[guiEdge.outputNode().nodeData]
            targetParameterId = guiEdge.outputPort().name
            contextManager.pomsetBuilder().connect(
                pomset,
                sourceNode, sourceParameterId,
                targetNode, targetParameterId)
            
        self.emit(
            SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return



    def OnDuplicateNodes(self, nodes):

        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        nodeMap = {}
        for guiNode in nodes:
            dataNode = guiNode.nodeData
            copyDataNode = contextManager.pomsetBuilder().copyNode(
                pomset, dataNode)
            copyDataNode.name('copy of ' + dataNode.name())

            nodeMap[dataNode] = copyDataNode
            pass

        self.emit(SIGNAL("OnNodesDuplicated(PyQt_PyObject)"), nodeMap)
        return nodeMap


    def OnDeleteSelection(self, selection):
        import pomsets_app.gui.graph as GraphModule
        nodes = [x for x in selection if isinstance(x, GraphModule.Node)]
        edges = [x for x in selection if isinstance(x, GraphModule.Edge)]

        # first remove the edges
        self.OnDeleteEdges(edges)

        # then remove the nodes
        self.OnDeleteNodes(nodes)
        return

    def OnDeleteEdges(self, edges):
        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        for guiEdge in edges:
            sourceNode = guiEdge.inputNode().nodeData
            sourceParameterId = guiEdge.inputPort().name
            targetNode = guiEdge.outputNode().nodeData
            targetParameterId = guiEdge.outputPort().name
            contextManager.pomsetBuilder().disconnect(
                pomset, 
                sourceNode, sourceParameterId,
                targetNode, targetParameterId
                )
            pass

        self.emit(
            SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return


    def OnDeleteNodes(self, nodes):
        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        for guiNode in nodes:
            dataNode = guiNode.nodeData
            contextManager.pomsetBuilder().removeNode(
                pomset, dataNode)
        self.emit(SIGNAL("OnNodesDeleted()"))
        return


    def OnCreateNode(self, pomsetContext, pomsetReference, 
                     definitionToReference, pos, name=None):

        contextManager = self.app().contextManager()

        if name is None:
            # create a node
            # set its position
            name = '%s_%s' % (definitionToReference.name() or "anonymous",
                              uuid.uuid4().hex[:3])

        node = contextManager.pomsetBuilder().createNewNode(
            pomsetReference.definitionToReference(), 
            name=name,
            definitionToReference=definitionToReference)

        pomsetContext.isModified(True)

        # emit a signal
        # so that the other widgets
        # will receive it and update accordingly
        self.emit(SIGNAL("OnNodeCreated(PyQt_PyObject,QPoint)"), node, pos)

        return


    def OnMoveDefinitionToLibrary(self, pomsetContext):
        contextManager = self.app().contextManager()
        persistentLibrary = contextManager.persistentLibrary()
        transientLibrary = contextManager.transientLibrary()
        definition = pomsetContext.pomset()

        # backup the existing library
        AppUtilsModule.backupLibrary(
            persistentLibrary.bootstrapLoaderDefinitionsDir())
        
        definitionFilter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdentityFilter(definition)
        )
        
        # remove from transient library
        transientLibrary.removeDefinition(definitionFilter)

        # define the url for the definition
        definitionPath = definition.id() + '.pomset'
        definition.url(definitionPath)
        outputDir = persistentLibrary.bootstrapLoaderDefinitionsDir()
        
        # save out the definition
        ContextModule.pickleDefinition(
            os.path.join(outputDir, definitionPath), definition)
        
        # add to persistent library
        definition.isLibraryDefinition(True)
        persistentLibrary.addPomsetContext(pomsetContext)
        
        # save out the bootstrap loader pomset
        # that will load this definition on the next startup
        persistentLibrary.saveBootstrapLoaderPomset()
        
        # update GUI
        self.emit(SIGNAL("OnPomsetMoved()"))

        return


    def OnDisplayDefinition(self, pomsetContext,
                            pomsetReferencePath, allowEdit=False):

        reference = pomsetReferencePath[-1]
        definitionToReference = reference.definitionToReference()

        if definitionToReference.isAtomic():
            return self.OnDisplayAtomicDefinition(
                pomsetContext, pomsetReferencePath,
                allowEdit=allowEdit)

        return self.OnDisplayNestDefinition(
            pomsetContext, pomsetReferencePath, allowEdit=allowEdit)


    def OnDisplayNestDefinition(self, pomsetContext,
                                pomsetReferencePath, allowEdit=False):

        # TODO:
        # need to be able to specify whether a nest pomset is editable
        #pomsetContext, pomsetReferencePath, isEditable = \
        #    contextManager.currentDisplayedPomsetInfo()

        self.displayPomset(pomsetContext, pomsetReferencePath=pomsetReferencePath)
        return


    def OnDisplayAtomicDefinition(self, pomsetContext, 
                                  pomsetReferencePath, allowEdit=False):

        reference = pomsetReferencePath[-1]

        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.edit_definition.widget as WidgetModule
        import pomsets_app.gui.qt.edit_definition.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.shouldAllowEdit(allowEdit)
        controller.contextManager(contextManager)
        controller.pomsetContext(pomsetContext)
        controller.definition(reference.definitionToReference())

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnEditParameterBinding(self, pomsetContext, definition, parameter):
        editTaskController = self.OnEditTask(pomsetContext, definition)
        controller = editTaskController.editParameterBinding(parameter)
        return controller


    def OnEditTask(self, pomsetContext, definition):
        """
        The task is actually the ReferenceDefinition
        from which the task will be built
        """

        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.edit_task.widget as WidgetModule
        import pomsets_app.gui.qt.edit_task.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)
        controller.pomsetContext(pomsetContext)
        controller.definition(definition)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return controller


    def OnDefinitionRemoved(self, library, pomsetContext):
        self.emit(SIGNAL("OnPomsetClosed(PyQt_PyObject)"), 
                  pomsetContext)
        return

    def OnCloudPoolConfiguration(self):

        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.cloud_pool.widget as WidgetModule
        import pomsets_app.gui.qt.cloud_pool.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnEuca2oolsConfiguration(self):
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.euca2ools_configuration.widget as WidgetModule
        import pomsets_app.gui.qt.euca2ools_configuration.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnHadoopConfiguration(self):
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.hadoop_configuration.widget as WidgetModule
        import pomsets_app.gui.qt.hadoop_configuration.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnExecutePreview(self):
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.task_preview.widget as WidgetModule
        import pomsets_app.gui.qt.task_preview.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnExecuteLocal(self):
        self.OnExecutePomset(AutomatonModule.ID_EXECUTE_LOCAL)
        return
    

    def OnExecuteEucalyptus(self):
        self.OnExecutePomset(AutomatonModule.ID_EXECUTE_EUCA2OOLS)
        return


    def OnAboutApplication(self):
        # TODO:
        # should load from file?

        text = """
<body>
<strong>%s</strong> %s<br/>""" % (str(self.app().applicationVersion()),
                                  str(self.app().applicationName())) + \
        u"\u00A9 2010 michael j pan"+ \
        """<br/>
</body>
"""
        informativeText = """
<body>
<a href="http://pomsets.org">pomsets</a> is workflow management <br/>
for your cloud<br/>
<br/>
<a href="http://pomsets.org/License">License</a><br/>
</body>
"""
        msgBox = QtGui.QMessageBox(self)
        msgBox.setTextFormat(Qt.RichText)
        msgBox.setText(text)
        msgBox.setInformativeText(informativeText)
        msgBox.exec_()
        return
        

    def OnExitApplication(self):

        modifiedContexts = []

        contextManager = self.app().contextManager()


        mainSplitterWidget = self.centralWidget()
        librarySplitterWidget = mainSplitterWidget.widget(0)
        toolboxWidget = librarySplitterWidget.widget(1)
        transientLibraryWidget = toolboxWidget.widget(0)
        activeLibraryWidget = librarySplitterWidget.widget(0)

        for widget in [activeLibraryWidget, transientLibraryWidget]:
            definitionTable = widget.pomsetLibrary().definitionTable()
            filter = FilterModule.constructAndFilter()
            filter.addFilter(widget.pomsetFilter())
            filter.addFilter(RelationalModule.ColumnValueFilter(
                    'context',
                    FilterModule.ObjectKeyMatchesFilter(
                        FilterModule.IdentityFilter(True),
                        keyFunction = lambda x: x.isModified())))
            modifiedContexts.extend(
                RelationalModule.Table.reduceRetrieve(
                    definitionTable,
                    filter, ['context'], []))
            pass

        canExit = len(modifiedContexts) is 0

        if not canExit:
            msgBox = QtGui.QMessageBox(self)
            msgBox.setText("There are unsaved, modified pomsets.")
            msgBox.setInformativeText("Discard changes or cancel exit?")
            msgBox.setStandardButtons(
                QtGui.QMessageBox.Discard | 
                QtGui.QMessageBox.Cancel)
            msgBox.setDefaultButton(QtGui.QMessageBox.Cancel)
            ret = msgBox.exec_()
            if ret == QtGui.QMessageBox.Discard:
                canExit = True
            
        if canExit:
            sys.exit(0)

        return
    

    def OnExecutePomset(self, executionEnvId):

        contextManager = self.app().contextManager()

        # TODO:
        # need to check and verify
        # that the specified threadpool
        # has at least one allocated thread
        # if not, should query user

        try:

            commandBuilderMap = contextManager.getDefaultCommandBuilderMap()
            executeEnvironmentMap = contextManager.getDefaultExecuteEnvironmentMap()
            
            requestKwds = contextManager.generateRequestKwds(
                executeEnvironmentId = executionEnvId,
                executeEnvironmentMap = executeEnvironmentMap,
                commandBuilderMap = commandBuilderMap
                )


            contextManager.executePomset(
                pomset=contextManager.activePomsetContext().reference(), 
                requestKwds=requestKwds,
                shouldWait=False)

        except Exception, e:
            import traceback
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
            stackTrace = \
                traceback.format_exception(exceptionType, exceptionValue,
                                           exceptionTraceback)
            logging.error("Error executing pomset >> %s" % stackTrace)

            # TODO:
            # show an error box to the user
            messageBox = QtGui.QMessageBox(self)
            # logging.error("Error executing pomset >> %s" % e)
            # messageBox.showMessage("Error executing pomset >> %s" % e)
            messageBox.setText('Execution error')
            messageBox.setInformativeText(str(e))
            messageBox.setIcon(QtGui.QMessageBox.Critical)
            messageBox.setStandardButtons(
                QtGui.QMessageBox.Ok)
            ret = messageBox.exec_()

            pass
        finally:
            # TODO:
            # update the execute menu items
            # i.e. disable execution for this pomset
            # now that execution has happened
            pass
            

        return


    # END class Frame
    pass




import wx
import pomsets_app.gui.widget as WidgetModule

class MessageDialog(WidgetModule.Dialog):

    XRC_NAME = 'TaskOutputMessagesDialog'

    WIDGET_BINDINGS = WidgetModule.Dialog.WIDGET_BINDINGS + [
        ('treectrlTasks', 'OnTaskSelected', wx.EVT_TREE_SEL_CHANGED)
    ]

    def __init__(self, xmlResource, parent=None):

        WidgetModule.Dialog.__init__(
            self, 
            xmlResource.LoadDialog(parent, MessageDialog.XRC_NAME))

        self.SetSize(wx.Size(575, 380))
        self.SetAutoLayout(True)

        treeCtrl = self.getWidget('treectrlTasks')
        treeCtrl.SetMinSize(wx.Size(150, 300))

        textOutputMessage = self.getWidget('textOutputMessage')
        textOutputMessage.SetMinSize(wx.Size(400,300))

        # okButton = wx.xrc.XRCCTRL(self, 'wxID_OK')
        # self.Bind(wx.EVT_BUTTON, self.OnButtonOk, okButton)
        # treeCtrl.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnTaskSelected)

        self._taskMessageMap = {}

        return


    def populateTaskTree(self, task):

        tree = self.getWidget('treectrlTasks')

        root = tree.AddRoot('Root')

        taskItemMap = {}
        taskParentItemMap = {}
        taskNameMap ={}

        tasksToAdd = [task]
        taskParentItemMap[task] = root
        taskNameMap[task] = task.definition().id()

        while len(tasksToAdd):
            task = tasksToAdd.pop(0)
            itemData = wx.TreeItemData()
            itemData.SetData(task)
            taskItem = tree.AppendItem(
                # taskParentItemMap[task], str(task), data=itemData)
                taskParentItemMap[task], 
                taskNameMap.get(task, str(task)), data=itemData
            )
            taskItemMap[task] = taskItem

            import pomsets.task as TaskModule
            if isinstance(task, TaskModule.CompositeTask):

                # get all the child tasks
                for index, childTask in enumerate(task.getChildTasks()):
                    taskParentItemMap[childTask] = taskItem
                    taskNameMap[childTask] = str(index)
                    tasksToAdd.append(childTask)
                    pass

                pass
            pass

        return

    def updateMessageBox(self):

        task = self._selectedTask

        messageBox = self.getWidget('textOutputMessage')

        value = self._taskMessageMap.get(task, None)
        if not value:
            request = task.workRequest()
            stream = request.kwds.get('process message output', None)

            lines = []

            executedCommand = request.kwds.get('executed command', None)
            if executedCommand:
                lines.append('# executed command\n')
                lines.append('# %s \n' % ' '.join(executedCommand))
                lines.append('\n')

            if stream is not None:
                lines.extend(stream.readlines())

            if request.kwds.get('exception stack trace', None) is not None:
                exceptionStackTrace = request.kwds.get('exception stack trace')
                lines.extend(exceptionStackTrace)

            value = ''.join(lines)
            self._taskMessageMap[task] = value
            pass

        if value:
            messageBox.SetValue(value)
            #print "messageBox class >> %s" % messageBox.__class__
            #page = "<html>%s</html>" % value
            #messageBox.SetPage(page)

        return

    def OnTaskSelected(self, event):

        tree = self.getWidget('treectrlTasks')

        treeItem = event.GetItem()
        itemData = tree.GetItemData(treeItem)
        task = itemData.GetData()

        self._selectedTask = task
        self.updateMessageBox()

        return


    pass




class ParameterValuesDialog(WidgetModule.Dialog):

    XRC_NAME = 'ParameterValuesDialog'
    WIDGET_BINDINGS = WidgetModule.Dialog.WIDGET_BINDINGS + [
        ('wxID_APPLY', 'OnButtonApply', wx.EVT_BUTTON)
    ]

    def __init__(self, xmlResource, parent=None):
        WidgetModule.Dialog.__init__(
            self, 
            xmlResource.LoadDialog(parent, ParameterValuesDialog.XRC_NAME))
        # self.Fit()
        self.SetSize(wx.Size(475, 475))
        self.SetAutoLayout(True)

        textCtrl = self.getWidget('textParameterValue')
        textCtrl.SetMinSize( wx.Size( 300,250 ) )
        staticText = self.getWidget('labelParameterDescription')
        staticText.SetMinSize( wx.Size( 300,100 ) );
        listBox = self.getWidget('listParameters')
        listBox.SetMinSize( wx.Size( 150,350 ) )

        self.Bind(wx.EVT_LISTBOX, self.OnSelect)



        return


    def dataNode(self, value=None):
        if value is not None:
            self._dataNode = value
        return self._dataNode

    def selectedParameter(self, value=None):
        if value is not None:
            self._selectedParameter = value
        return self._selectedParameter


    def OnButtonApply(self, event):
        (dataNode, parameter) = \
         self.dataNode().getParameterToEdit(self.selectedParameter())
        textCtrl = self.getWidget('textParameterValue')
        text = textCtrl.GetValue()
        values = text.split("\n")

        # the values are to be set on the blackboard parameter
        parameterId = parameter.id()
        logging.debug('binding parameter %s to value "%s" on dataNode %s' % 
                      (parameterId, values, dataNode))
        dataNode.setParameterBinding(
            parameterId, values)

        # file staging and parameter sweep are to be set on the 
        # actual parameter
        parameter = self.selectedParameter()

        # set the value for file staging
        checkboxFileStaging = self.getWidget('checkboxFileStaging')
        value = checkboxFileStaging.GetValue()
        self.dataNode().parameterStagingRequired(
            parameter, value)

        # set the value for parameter sweep
        checkboxParameterSweep = self.getWidget('checkboxParameterSweep')
        value = checkboxParameterSweep.GetValue()
        logging.debug("in ParameterValuesDialgo.OnButtonApply")
        self.dataNode().isParameterSweep(
            parameter, value
        )

        return


    def OnSelect(self, event):

        listBox = self.getWidget('listParameters')
        selectionIndex = listBox.GetSelection()

        # this should be the parameter object
        # event.ClientData
        self.selectedParameter(event.ClientData.id())


        self.updateValues()

        return

    def updateValues(self):
        # could probably combine populateTextCtrl and SetParameterName
        # by calling something like setParameter(parameter)
        self.populateTextCtrl()
        self.setParameterName()

        parameterId = self.selectedParameter()
        parameter = self.dataNode().getParameter(parameterId)

        # enable the checkbox only if the value is supposed to be a file
        checkboxFileStaging = self.getWidget('checkboxFileStaging')
        if parameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT) or\
           parameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE):
            checkboxFileStaging.Enable(True)

            value = self.dataNode().parameterStagingRequired(parameterId)
            checkboxFileStaging.SetValue(value)


        else:
            checkboxFileStaging.SetValue(False)
            checkboxFileStaging.Enable(False)

        # update the value for parameter sweep
        checkboxParameterSweep = self.getWidget('checkboxParameterSweep')
        value = self.dataNode().isParameterSweep(parameterId)
        checkboxParameterSweep.SetValue(value)

        return

    def populateParameterList(self):
        """
        fill in the values for the list box
        """

        listBox = self.getWidget('listParameters')

        filter = FilterModule.ObjectKeyMatchesFilter(
            keyFunction=lambda x: x.portType(),
            filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_TEMPORAL)
        )
        notFilter = FilterModule.constructNotFilter()
        notFilter.addFilter(filter)

        for index, parameter in enumerate(
            self.dataNode().getParametersByFilter(notFilter)):

            listBox.Insert(parameter.name(),
                           index,
                           parameter)
            pass

        return

    def selectParameterInListBox(self):
        parameterName = self.selectedParameter()
        listBox = self.getWidget('listParameters')
        listBox.SetStringSelection(parameterName)
        return

    def setParameterName(self):
        staticText = self.getWidget('labelParameterDescription')
        staticText.SetLabel(self.selectedParameter())
        return

    def populateTextCtrl(self):

        # first we have to get the actual dataNode and parameter
        (dataNode, parameter) = \
         self.dataNode().getParameterToEdit(self.selectedParameter())

        values = ''
        try:
            values = '\n'.join(dataNode.getParameterBinding(parameter.id()))
        except KeyError, e:
            logging.error(e)
            pass

        textCtrl = self.getWidget('textParameterValue')
        textCtrl.SetValue(values)
        return


    # END class ParameterValuesDialog
    pass


class ExecutionPreviewDialog(WidgetModule.Dialog):

    ID_EXECUTIONPREVIEW = wx.NewId()

    XRC_NAME = 'ExecutionPreviewDialog'

    WIDGET_BINDINGS = WidgetModule.Dialog.WIDGET_BINDINGS + []

    def __init__(self, xmlResource, parent=None):
        WidgetModule.Dialog.__init__(
            self, 
            xmlResource.LoadDialog(parent, ExecutionPreviewDialog.XRC_NAME))
        self.SetSize(wx.Size(520, 375))
        self.SetAutoLayout(True)


        textCtrl = self.getWidget('textCtrl')
        textCtrl.SetMinSize( wx.Size( 500,300 ) )


        # okButton = wx.xrc.XRCCTRL(self, 'wxID_OK')
        # self.Bind(wx.EVT_BUTTON, self.OnButtonOk, okButton)

        return


    def populateTextCtrl(self, value):
        # TODO:
        # will have to hande the case where nodes are selected

        textCtrl = self.getWidget('textCtrl')
        textCtrl.SetValue(value)

        return


    # END ExecutionPreviewDialog
    pass



import wx
import wx.grid

import pypatterns.command as CommandModule

import pypatterns.filter as FilterModule

import pypatterns.relational as RelationalModule
import pypatterns.relational.commands as RelationalCommandModule

# possible notification messages
#['GRIDTABLE_NOTIFY_COLS_APPENDED', 'GRIDTABLE_NOTIFY_COLS_DELETED',
#'GRIDTABLE_NOTIFY_COLS_INSERTED', 'GRIDTABLE_NOTIFY_ROWS_APPENDED',
#'GRIDTABLE_NOTIFY_ROWS_DELETED', 'GRIDTABLE_NOTIFY_ROWS_INSERTED',
#'GRIDTABLE_REQUEST_VIEW_GET_VALUES', 'GRIDTABLE_REQUEST_VIEW_SEND_VALUES']


class NotifyGridCommand(CommandModule.Command):
    
    def __init__(self, grid, executeIfUndo=False):
        CommandModule.Command.__init__(self)
        self._grid = grid
        self._executeIfUndo = executeIfUndo
        return

    
        
    def execute(self):
        if self._executeIfUndo:
            return

        msg = self.buildMessageForExecute()

        self._grid.ProcessTableMessage(msg)
        
        return
    
    def unexecute(self):
        if not self._executeIfUndo:
            return
        
        msg = self.buildMessageForUnexecute()
        
        self._grid.ProcessTableMessage(msg)
        
        return
    
    # END class NotifyGridCommand
    pass


class NotifyValueChangedCommand(NotifyGridCommand):

    def buildMessageForExecute(self):
        msg = wx.grid.GridTableMessage(
            self._grid.GetTable(),
            wx.grid.GRIDTABLE_REQUEST_VIEW_GET_VALUES
        )
        return msg

    def buildMessageForUnexecute(self):
        
        # it's the same thing to notify a value changed
        # because of an undo
        return self.buildMessageForExecute()
    
    pass
    
    
class NotifyRowAddedCommand(NotifyGridCommand):

    
    def buildMessageForExecute(self):
        # tell the grid we've added a row
        msg = wx.grid.GridTableMessage(
            self._grid.GetTable(),
            wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED,
            1
        )
        return msg
    
    def buildMessageForUnexecute(self):
        # tell the grid we've added a row
        msg = wx.grid.GridTableMessage(
            self._grid.GetTable(),
            wx.grid.GRIDTABLE_NOTIFY_ROWS_DELETED,
            1
        )
        return msg
    

    # END class NotifyRowAddedCommand
    pass


class NotifyGridCommandBuilder(CommandModule.CommandBuilder):
    """
    command builder for the notify grid command classes
    """
    
    def __init__(self, command, notifyCommandClass, grid):
        CommandModule.CommandBuilder.__init__(self, command)
        self._notifyCommandClass = notifyCommandClass
        self._grid = grid
        return
    
    def addCommandsPreExecute(self, commands):

        commands.append(
            self._notifyCommandClass(self._grid, executeIfUndo=True)
        )
        
        return

    def addCommandsPostExecute(self, commands):
        commands.append(
            self._notifyCommandClass(self._grid, executeIfUndo=False)
        )
        return

    # END class NotifyGridCommandBuilder
    pass

    
class UpdateGridCellValueCommandBuilder(CommandModule.CommandBuilder):

    def __init__(self, command, grid, rowFilter, column, value):
        """
        @param grid the wx.Grid
        """
        CommandModule.CommandBuilder.__init__(self, command)
        self._grid = grid
        self._column = column
        self._rowFilter = rowFilter
        self._value = value
        return

    def getValueToSet(self):
        return self._value
    
    def addCommandsPostExecute(self, commands):
        
        dataModel = self._grid.GetTable().dataModel()
        for row in dataModel.retrieveForModification(self._rowFilter):
            command = CommandModule.CompositeCommand()
            setValueCommand = RelationalCommandModule.SetColumnValueCommand(
                row, self._column, self.getValueToSet()
            )
            
            command.addCommand(setValueCommand)
            command.addCommandBuilder(
                setValueCommand,
                NotifyGridCommandBuilder(
                    setValueCommand, NotifyValueChangedCommand, self._grid
                )
            )
            
            commands.append(command)
            pass
        
        return

    # END class UpdateGridCellValueCommandBuilder
    pass

class SetVmThreadCommandBuilder(UpdateGridCellValueCommandBuilder):
    def __init__(self, command, grid, vmInstance):
        UpdateGridCellValueCommandBuilder.__init__(
            self, command, grid, 
            RelationalModule.ColumnValueFilter(
                'Instance ID',
                FilterModule.EquivalenceFilter(vmInstance)
            ),
            'Thread',
            None
        )
        return
    
    def getValueToSet(self):
        return self._command._worker
    
    # END class SetVmThreadCommandBuilder
    pass


class UnsetVmThreadCommandBuilder(UpdateGridCellValueCommandBuilder):
    def __init__(self, command, grid):
        UpdateGridCellValueCommandBuilder.__init__(
            self, command, grid, 
            RelationalModule.ColumnValueFilter(
                'Instance ID',
                FilterModule.EquivalenceFilter(command._vmInstance)
            ),
            'Thread',
            None
        )
        return
    
    def getValueToSet(self):
        return RelationalModule.Table.NULL
    
    # END class UnsetVmThreadCommandBuilder
    pass




import wx
import wx.glcanvas 

import pomsets_app.gui.canvas as CanvasModule

class Canvas(wx.glcanvas.GLCanvas, CanvasModule.Canvas):


    def __init__(self, parent, *args, **kwds):

        attribList = (
            wx.glcanvas.WX_GL_DOUBLEBUFFER,
            wx.glcanvas.WX_GL_RGBA, 0
        )
        wx.glcanvas.GLCanvas.__init__(
            self, parent, -1, attribList = attribList,
            *args, **kwds)

        CanvasModule.Canvas.__init__(self, *args, **kwds)
        return


    def Render(self, dc):
        if self._frameIsInProcessOfShowing:
            return

        """
        copied from zgl_graphdrawer.wx_canvas.Canvas
        """

        self.SetCurrent()

        self.displayFunc()

        self.SwapBuffers()
        return


    def forceRedraw(self):
        """
        copied from zgl_graphdrawer.wx_canvas.Canvas
        """
        dc = wx.PaintDC(self)
        self.Render(dc)
        return

    def initializeEventHandlers(self, contextManager=None):
        self.contextManager(contextManager)

        for key, default in [
            ('mouse event handler', EventModule.MouseEventHandler),
            ('key event handler', EventModule.KeyEventHandler),
            ('canvas event handler', EventModule.CanvasEventHandler),
            ]:

            classObject = contextManager.app().getResourceValue(
                "%s class" % key, default=default)

            eventHandler = classObject(self)
            eventHandler.bindEvents()
            self.PushEventHandler(eventHandler)

            # TODO:
            # remove this once the event handler mechanism works
            contextManager.app().setResourceValue(
                key, eventHandler)

            pass

        return


    def initializeTimer(self):
        """
        copied from zgl_graphdrawer.wx_canvas.Canvas
        """

        self.TIMER_ID = 100  # pick a number

        # message will be sent to the parent
        self.timer = wx.Timer(self.GetParent(), self.TIMER_ID)  
        self.timer.Start(100)  # x100 milliseconds

        # TODO:
        # figure out how to have multiple listeners for the time event
        # currently only the CanvasEventHandler is handling it

        return

    def drawMetaInfo(self):
        
        CanvasModule.Canvas.drawMetaInfo(self)

        # draw breadcrumbs
        x = 30
        y = self.height - 45
        glColor3f(1.0, 1.0, 1.0)
        zglUtils.drawBitmapText(
            self.breadCrumbs(),
            x, y)
        
        return

    def drawFPS(self, x, y):
        if hasattr(self, "last_t"):
            self.fps = 1.0/float(time.time()-self.last_t)
            glColor3f(1.0, 1.0, 1.0)

            zglUtils.drawBitmapText(
                "FPS: %f" % (self.fps), 
                x, y,
                font=GLUT_BITMAP_HELVETICA_10)
        self.last_t = time.time()

        return


    # END class Canvas
    pass



import os
import sys

APP_ROOT = os.getenv('APP_ROOT')

sys.path.insert(0, '%s/pomsets-gui/src' % APP_ROOT)
sys.path.insert(0, '%s/pomsets-core/src' % APP_ROOT)
sys.path.insert(0, '%s/cloudpool/src' % APP_ROOT)
sys.path.insert(0, '%s/pypatterns/src' % APP_ROOT)

from pomsets_qt import *

if __name__=="__main__":
    main()



import copy
import functools
import logging
import os
import time

import simplejson as ConfigModule

import currypy 

import pypatterns.command as CommandModule

import pomsets.automaton as AutomatonModule
import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule
import pypatterns.relational.commands as RelationalCommandModule

import pomsets.task as TaskModule

import cloudpool.shell as ShellModule

import pomsets_app.controller as CloudModule


ID_EXECUTE_LOCAL = 1
ID_EXECUTE_REMOTE = 2
ID_EXECUTE_EUCA2OOLS = 3

class Automaton(AutomatonModule.Automaton):

    ATTRIBUTES = AutomatonModule.Automaton.ATTRIBUTES + [
        'cloudControllerTable',
        'virtualMachineTable',
        'eventHandlers',
        'shouldProcessEndedTasks',
        'remoteExecuteCredentials',
        'threadPoolTable',
        'threadTable',
        'executeEnvironmentTable',
        'hadoopConfigurations',
        'otherConfigurations'
    ]

    COLUMNS_VIRTUALMACHINE = [
        'API', 'Instance ID', 'Service', 'Reservation ID', 'Image ID', 'IP', 
        'Enabled', 'Key name', 
    ]

    COLUMN_EXECUTEENVIRONMENTS = [
        'Instance ID', 
        'Execute environment'
    ]

    COLUMNS_THREADS = [
        'Thread', 'Execute environment', 'Assigned task'
    ]

    COLUMNS_CLOUDCONTROLLERCREDENTIALS = [
        'serviceName', 'serviceAPI', 'values']


    @staticmethod
    def doTask(*args, **kwds):
        AutomatonModule.Automaton.doTask(*args, **kwds)
        return


    @staticmethod
    def compositeTaskCompleteCallback(request, result):
        """
        this is meant to be called after the composite task
        has initialized all the minimal children,
        not after all maximal children have completed execution
        """
        # TODO:
        # update the gui

        AutomatonModule.Automaton.compositeTaskCompleteCallback(request, result)

        return


    @staticmethod
    def atomicTaskCompleteCallback(request, result):

        # TODO:
        # update the gui

        AutomatonModule.Automaton.atomicTaskCompleteCallback(request, result)

        return



    def __init__(self):

        AutomatonModule.Automaton.__init__(self)

        logging.debug("initializing automaton")

        self.virtualMachineTable(
            RelationalModule.createTable(
                'virtual machines',
                Automaton.COLUMNS_VIRTUALMACHINE
            )
        )

        self.executeEnvironmentTable(
            RelationalModule.createTable(
                'execute environments',
                Automaton.COLUMN_EXECUTEENVIRONMENTS
            )
        )

        self.threadTable(
            RelationalModule.createTable(
                'threads',
                Automaton.COLUMNS_THREADS
            )
        )

        self.eventHandlers(set([]))


        table = RelationalModule.createTable(
            'clouds',
            Automaton.COLUMNS_CLOUDCONTROLLERCREDENTIALS
        )
        self.cloudControllerTable(table)

        table = RelationalModule.createTable(
            'thread pools',
            ['id', 'threadPool']
        )
        self.threadPoolTable(table)

        self.remoteExecuteCredentials([])
        self.hadoopConfigurations({})

        pass


    def saveCloudControllerCredentials(self):
        raise NotImplementedError(
            'have not implemented saving cloud credentials to file')

    def loadConfig(self, config):
        self.loadCloudControllerCredentials(config)
        self.loadRemoteExecuteCredentials(config)
        self.loadHadoopConfigurationValues(config)
        self.loadOtherConfigurations(config)
        return

    def addConfigToSave(self, configData):
        self.addCloudControllerCredentialsToConfig(configData)
        self.addRemoteExecuteCredentialsToConfig(configData)
        self.addHadoopConfigurationsToConfig(configData)
        self.addOtherConfigurationsToSave(configData)
        return

    def loadOtherConfigurations(self, config):
        configCopy = copy.copy(config)
        
        for keyToRemove in ['hadoop configurations',
                            'remote execute credentials',
                            'cloud controller credentials']:
            if keyToRemove in configCopy:
                del configCopy[keyToRemove]

        self.otherConfigurations(configCopy)
        return

    #def addOtherConfigurationsToSave(self, key, configData):
    #    if not key in self.otherConfigurations():
    #        self.otherConfigurations()[key] = []
    #    self.otherConfigurations()[key].append(configData)
    #    return
    def addOtherConfigurationsToSave(self, configData):
        configData.update(self.otherConfigurations())
        return

    def addCloudControllerCredentialsToConfig(self, config):

        credentials = []
        table = self.cloudControllerTable()
        tableColumns = table.columns()
        configKeys = [x for x in tableColumns]
        configKeys[configKeys.index('serviceName')] = 'service name'
        configKeys[configKeys.index('serviceAPI')] = 'service API'
        for row in table.retrieve(columns=tableColumns):
            credentials.append(dict(zip(configKeys, row)))
            pass
        config['cloud controller credentials'] = credentials
        return

    def addRemoteExecuteCredentialsToConfig(self, config):
        config['remote execute credentials'] = self.remoteExecuteCredentials()
        return

    def loadRemoteExecuteCredentials(self, config):
        for credential in config['remote execute credentials']:
            #hostname = credential['hostname']
            #user = credential['user']
            #keyfile = credential['keyfile']
            # self.addRemoteExecuteCredential(hostname, user, keyfile)
            self.remoteExecuteCredentials().append(credential)
            pass
        return


    def loadCloudControllerCredentials(self, config):

        for credential in config['cloud controller credentials']:
            serviceName = credential['service name']
            serviceAPI = credential['service API']
            rawValues = credential['values']

            values = dict([(str(x), str(rawValues[x])) for x in rawValues.keys()])
            self.addCloudControllerCredential(serviceName, serviceAPI, **values)

            pass

        return


    def getCloudControllerCredentialsForAPI(self, serviceAPI):
        filter = RelationalModule.ColumnValueFilter(
            'serviceAPI',
            FilterModule.EquivalenceFilter(serviceAPI)
        )
        values = [x for x in self.getCloudControllerCredentials(filter)]
        if len(values) is not 1:
            raise KeyError("not implemented to handle %s values" % values)

        return values[0][0]


    def getCloudControllerCredentials(self, filter=None, columns=None):

        if columns is None:
            columns = ['values']

        if filter is None:
            filter = FilterModule.TRUE_FILTER

        for x in self.cloudControllerTable().retrieve(
            filter = filter,
            columns = columns):
            yield x
        raise StopIteration


    def addCloudControllerCredential(self, serviceName, serviceAPI, **kwds):
        """
        Possible values:
        * (Amazon, EC2, ...)
        * (Amazon, Euca2ools, ...)
        * (Eucalyptus, Euca2ools, ...)
        * (Rackspace, Rackspace, ...)
        """

        row = self.cloudControllerTable().addRow()
        row.setColumn('serviceName', serviceName)
        row.setColumn('serviceAPI', serviceAPI)
        row.setColumn('values', kwds)

        return

    def loadHadoopConfigurationValues(self, config):
        for configuration in config['hadoop configurations']:
            hostname = configuration.get('hostname', 'localhost')
            self.hadoopConfigurations()[hostname] = configuration
            pass
        return

    def addHadoopConfiguration(self, hadoopConfig):
        self.hadoopConfigurations()['hostname'] = hadoopConfig
        return

    def addHadoopConfigurationsToConfig(self, config):
        config['hadoop configurations'] = self.hadoopConfigurations().values()
        return

    def openConnection(self, name):

        # TODO:
        # should retrieve the value from the table
        # based upon the input parameter name
        kwds = CloudModule.defaultEC2Values()

        import boto
        import boto.ec2.regioninfo

        connection = boto.connect_ec2(
            aws_access_key_id=kwds[CloudModule.KEY_CONTROLLER_ACCESS_KEY],
            aws_secret_access_key=kwds[CloudModule.KEY_CONTROLLER_SECRET_KEY],
            is_secure=kwds['isSecure'],
            region=boto.ec2.regioninfo.RegionInfo(
                None, 
                kwds[CloudModule.KEY_CONTROLLER_SERVICE_NAME],
                kwds['host']
                ),
            port=kwds['port'],
            path=kwds['path']
        )

        return connection


    def _terminateVmAndThread(self):

        if True:
            raise NotImplementedError('_terminateVmAndThread')

        # start a thread that will wait to 
        thread = CloudModule.Thread(
            functools.partial(
                self.threadpool().dismissWorker,
                callback=self._removeVmEntryOfThread
            )
        )
        thread.start()
        return


    def _removeVmEntryOfThread(self, workerThread):

        # get the image id for the thread
        theFilter = RelationalModule.ColumnValueFilter(
            'Thread',
            FilterModule.IdentityFilter(workerThread)
        )
        instanceIds = RelationalModule.Table.reduceRetrieve(
            self.virtualMachineTable(),
            theFilter,
            ['Instance ID']
        )
        if not len(instanceIds):
            raise ValueError('Could not find any VMs assigned to thread %s' % 
                             workerThread)

        # remove the entry from the database
        # connection = self.openConnection('eucalyptus')
        # instances = connection.terminate_instances(instanceIds)
        self.terminateVmInstanceUsingIds(instanceIds)

        # now remove the rows
        self.virtualMachineTable().removeRows(theFilter)

        return


    def terminateVmInstance(self, instance):
        id = instance.id
        return self.terminateVmInstanceUsingIds([id])

    def terminateVmInstanceUsingIds(self, instanceIds):

        # remove the entry from the database
        connection = self.openConnection('eucalyptus')
        terminatedInstances = connection.terminate_instances(instanceIds)

        return


    def provisionVM(self, name, keyPair, imageId):

        # TODO:
        # request the machine be provisioned
        # image_id = None
        # TODO: switch image_id to be configurable

        kernel_id = None
        ramdisk_id = None
        min_count = 1
        max_count = 1
        instance_type = 'm1.small' 
        group_names = [] 
        user_data = None
        addressing_type = None
        zone = None

        connection = self.openConnection(name)

        reservation = connection.run_instances(
            image_id = imageId,
            min_count = min_count,
            max_count = max_count,
            key_name = keyPair,
            security_groups = group_names,
            user_data = user_data,
            addressing_type = addressing_type,
            instance_type = instance_type,
            placement = zone,
            kernel_id = kernel_id,
            ramdisk_id = ramdisk_id)


        # TODO:
        # add an entry into the table

        numInstances = len(reservation.instances)
        if numInstances is not 1:
            raise NotImplementedError('expected 1 instance, got %s' % 
                                      numInstances)
        instance = reservation.instances[0]

        return instance


    def unassignExecuteEnvironmentFromVmInstance(self, 
                                                 vmInstance, 
                                                 executeEnvironment=None):

        filter = RelationalModule.ColumnValueFilter(
            'Instance ID',
            FilterModule.EquivalenceFilter(vmInstance)
        )

        # TODO:
        # should enable removal of only a single row
        self.executeEnvironmentTable().removeRows(filter)
        return


    def assignExecuteEnvironmentForVmInstance(self, serviceAPI, vmInstance):

        filter = RelationalModule.ColumnValueFilter(
            'Instance ID',
            FilterModule.EquivalenceFilter(vmInstance)
        )
        credentials = self.getCloudControllerCredentialsForAPI(serviceAPI)

        for host, key in self.virtualMachineTable().retrieve(filter, ['IP', 'Key name']):
            if not key == credentials['user key pair']:
                logging.debug('key pair does not match')
                continue

            # create the env
            env = ShellModule.SecureShell()
            env.user('root')
            env.keyfile(credentials['identity file'])
            env.hostname(host)

            # add to the execute environment table 
            row = self.executeEnvironmentTable().addRow()
            row.setColumn('Execute environment', env)
            row.setColumn('Instance ID', vmInstance)

            return env

        raise KeyError('no instance with id %s found' % vmInstance)


    def assignThreadForExecuteEnvironment(self, threadPoolID, executeEnvironment):

        threadpool = self.getThreadPoolUsingId(threadPoolID)

        workerThread = threadpool.assignWorker()
        workerThread.executeEnvironment(executeEnvironment)

        row = self.threadTable().addRow()
        row.setColumn('Thread', workerThread)
        row.setColumn('Execute environment', executeEnvironment)

        return workerThread

    def getServiceApiFromVmInstance(self, vmInstance):
        filter = RelationalModule.ColumnValueFilter(
            'Instance ID',
            FilterModule.EquivalenceFilter(vmInstance)
        )

        apis = RelationalModule.Table.reduceRetrieve(
            self.virtualMachineTable(),
            filter,
            ['API'],
            []
        )
        if len(apis) is not 1:
            raise NotImplementedError(
                'not implemented for a vmInstance to have %s APIs' % len(apis))

        return apis[0]


    def getExecuteEnvironmentsFromVmInstance(self, vmInstance):

        filter = RelationalModule.ColumnValueFilter(
            'Instance ID',
            FilterModule.EquivalenceFilter(vmInstance)
        )

        executeEnvironments = RelationalModule.Table.reduceRetrieve(
            self.executeEnvironmentTable(),
            filter,
            ['Execute environment'],
            []
        )
        return executeEnvironments


    def unassignThreadFromVmInstance(self, threadPoolId, instance):
        """
        # TODO:
        # should be able to pull threadPoolId from the row as well
        """
        executeEnvironments = self.getExecuteEnvironmentsFromVmInstance(instance)

        self.unassignThreadFromExecuteEnvironment(
            threadPoolId, executeEnvironments[0])
        return


    def getThreadsFromExecuteEnvironment(self, executeEnvironment):
        filter = RelationalModule.ColumnValueFilter(
            'Execute environment',
            FilterModule.IdentityFilter(executeEnvironment)
        )
        threads = RelationalModule.Table.reduceRetrieve(
            self.threadTable(),
            filter,
            ['Thread'],
            []
        )
        return threads


    def unassignThreadFromExecuteEnvironment(self, threadPoolId, executeEnvironment):

        threads = self.getThreadsFromExecuteEnvironment(executeEnvironment)

        threadPool = self.getThreadPoolUsingId(threadPoolId)
        threadPool.dismissWorker(worker=threads[0])
        return


    def executeCommand(self, command):

        success = False
        try:
            self.commandManager().do(command)
            success = True
        except CommandModule.ExecutionError, e:

            # TODO:
            # notify the user that the command has failed
            pass

        return success


    def getThreadPoolUsingRequest(self, request):

        threadPoolId = request.kwds['execute environment id']
        return self.getThreadPoolUsingId(threadPoolId)


    def getThreadPoolUsingId(self, threadPoolId):

        filter = RelationalModule.ColumnValueFilter(
            'id',
            FilterModule.IdentityFilter(threadPoolId)
        )

        threadpools = RelationalModule.Table.reduceRetrieve(
            self.threadPoolTable(),
            filter,
            ['threadPool'],
            []
        )

        if len(threadpools) is 0:
            raise ValueError('no thread pool for id %s' % threadPoolId)
        if len(threadpools) is not 1:
            raise ValueError('too many thread pools for id %s' % threadPoolId)

        return threadpools[0]


    def setThreadPool(self, id, threadpool):
        row = self.threadPoolTable().addRow()
        row.setColumn('id', id)
        row.setColumn('threadPool', threadpool)
        return


    def enqueueRequest(self, request, shouldWait=True):
        """
        we don't want to call the superclass
        because currently the automaton in the pomset library
        can handle only one threadpool
        and waits on that threadpool when executing a request
        """

        # set the tasks' automaton as self
        task = request.kwds['task']
        task.automaton(self)

        if hasattr(self.contextManager(), 'canvas'):
            canvas = self.contextManager().canvas
            definition = task.definition()

            parentTask = task.parentTask()
            if parentTask and parentTask.definition() is definition:
                # do nothing
                # because this is a parameter sweep
                # and we don't want the status of the children
                # of overwrite the status of the parent
                # which is actually what's being shown to the user
                logging.debug("not configuring performer for child parameter sweep task %s for definition %s" % (task, definition.name()))
                pass
            elif canvas.hasGuiObjectForDataNode(definition):
                # setup the canvas gui object to listen for updates
                # on the execution status
                guiNode = canvas.getGuiObjectForDataNode(definition)
                guiNode.configurePerformers(task)

                logging.debug("configured performers for task %s (definition %s)" % (task, definition.name()))
                pass
            else:
                logging.debug("canvas does not have gui object for task %s's definition %s" % (task, definition.name()))
                pass
        else:
            logging.debug('context manager does not have reference to a canvas, most likely because we are executing the library bootstrap loader')

        logging.debug('enqueuing task %s' % task)
        AutomatonModule.Automaton.enqueueRequest(
            self, request, shouldWait=shouldWait)

        return


    def getPostExecuteCallbackFor(self, task):
        if isinstance(task, TaskModule.CompositeTask):
            return Automaton.compositeTaskCompleteCallback
        return Automaton.atomicTaskCompleteCallback



    def getExecuteTaskFunction(self, task):
        return Automaton.doTask


    def startProcessEndedTasksThread(self):

        thread = CloudModule.Thread(self.processEndedTasks)
        thread.start()
        return


    def processEndedTasks(self):
        """
        threadpool requires that we call wait()
        so that we can be notified of upates of execution status
        so we put this in a separate thread
        otherwise the GUI will block
        """

        # TODO:
        # add a check to verify that this is the only one running
        logging.debug('starting processing ended tasks')
        self.shouldProcessEndedTasks(True)
        while self.shouldProcessEndedTasks():

            threadpools = RelationalModule.Table.reduceRetrieve(
                self.threadPoolTable(),
                FilterModule.TRUE_FILTER,
                ['threadPool'],
                []
            )
            for threadpool in threadpools:
                threadpool.wait()

            time.sleep(1)
            pass
        logging.debug('ending processing ended tasks')
        return


    def addValuesToRequestContextForBootstrapLoader(self, requestContext, *args, **kwds):
        AutomatonModule.Automaton.addValuesToRequestContextForBootstrapLoader(
            self, requestContext, *args, **kwds)
        requestContext['execute environment id'] = ID_EXECUTE_LOCAL
        return


    # END class Automaton
    pass



import wx

ID_DIALOG_SELECTDEFINITION = wx.NewId()
ID_DIALOG_PARAMETERVALUES = wx.NewId()
ID_DIALOG_EXECUTIONPREVIEW = wx.NewId()
ID_DIALOG_TASKMESSAGES = wx.NewId()



class EvtHandler(wx.EvtHandler):
    
    WIDGET_BINDINGS = []
    
    def __init__(self, other):
        self.this = other.this
        self.thisown = 1
        del other.this
        self._setOORInfo(self)
        self.other = other
        
        self.bindWidgets()

        return
    
    
    def getWidget(self, widgetId):
        return wx.xrc.XRCCTRL(self, widgetId)
    
    def bindWidgets(self):
        for xrcId, nameOfFunctionToBind, eventId in self.__class__.WIDGET_BINDINGS:
            functionToBind = getattr(self, nameOfFunctionToBind)
            widget = self.getWidget(xrcId)
            widget.Bind(eventId, functionToBind)
            pass
        return
    
    pass


class Panel(wx.Panel, EvtHandler):

    WIDGET_BINDINGS = []
    
    def __init__(self, xmlResource, parent=None):
        EvtHandler.__init__(
            self, xmlResource.LoadPanel(parent, 'loadedDefinitionPanel'))
        return

    # END class Panel
    pass


class Dialog(wx.Dialog, EvtHandler):

    WIDGET_BINDINGS = [
        ('wxID_OK', 'OnButtonOk', wx.EVT_BUTTON)
    ]

    def __init__(self, other):
        EvtHandler.__init__(self, other)
        return

    def OnButtonOk(self, event):
        self.EndModal(wx.ID_OK)
        return

    pass



import logging
import os

from PyQt4.QtCore import *
from PyQt4 import QtGui

import pypatterns.filter as FilterModule

import pomsets.parameter as ParameterModule

import pomsets_app.utils as AppUtilsModule
import pomsets_app.gui.qt as QtModule

class Controller(QtModule.Controller, QtGui.QDialog):


    def __init__(self, *args, **kwds):
        QtGui.QDialog.__init__(self, *args, **kwds)
        QtModule.Controller.__init__(self)
        return

    def populate(self):

        widget = self.widget()
        contextManager = self.contextManager()

        controllerValues = {}
        try:
            controllerValues = contextManager.automaton().getCloudControllerCredentialsForAPI('euca2ools')
        except KeyError:
            pass

        textEditURL = widget.textEditURL
        textEditURL.setText(controllerValues.get('url', ''))

        textEditAccessKey = widget.textEditAccessKey
        textEditAccessKey.setText(controllerValues.get('access key', ''))

        textEditSecretKey = widget.textEditSecretKey
        textEditSecretKey.setText(controllerValues.get('secret key', ''))

        textEditKeyPair = widget.textEditKeyPair
        textEditKeyPair.setText(controllerValues.get('user key pair', ''))

        textEditIdentityFile = widget.textEditIdentityFile
        textEditIdentityFile.setText(controllerValues.get('identity file', ''))
        return


    @pyqtSlot(QtGui.QAbstractButton)
    def on_buttonBox_clicked(self, button):
        widget = self.widget()
        buttonRole = widget.buttonBox.buttonRole(button)
        if buttonRole == QtGui.QDialogButtonBox.ApplyRole:
            self.saveValues()
            pass
        if buttonRole == QtGui.QDialogButtonBox.AcceptRole:
            self.saveValues()
            self.accept()
            pass
        if buttonRole == QtGui.QDialogButtonBox.RejectRole:
            self.reject()
            pass
        return


    @pyqtSlot()
    def on_buttonBrowse_clicked(self):
        widget = self.widget()
        textEditIdentityFile = widget.textEditIdentityFile

        import user
        initialDir = user.home
        identityFile = str(textEditIdentityFile.text())
        if len(identityFile):
            index = identityFile.rfind(os.path.sep)
            if not index == -1:
                initialDir = identityFile[:index]
            pass

        fileName = QtGui.QFileDialog.getOpenFileName(
            self,
            QString("SSH identity file"), 
            QString(initialDir))
        if fileName and len(fileName):
            textEditIdentityFile.setText(fileName)
        return


    def saveValues(self):

        widget = self.widget()
        contextManager = self.contextManager()
        automaton = contextManager.automaton()

        shouldCreateNew = False
        controllerValues = {}
        try:
            controllerValues = automaton.getCloudControllerCredentialsForAPI('euca2ools')
        except KeyError, e:
            shouldCreateNew = True
            pass

        textEditKeyPair = widget.textEditKeyPair
        controllerValues['user key pair'] = str(textEditKeyPair.text())

        textEditURL = widget.textEditURL
        controllerValues['url'] = str(textEditURL.text())

        textEditAccessKey = widget.textEditAccessKey
        controllerValues['access key'] = str(textEditAccessKey.text())

        textEditSecretKey = widget.textEditSecretKey
        controllerValues['secret key'] = str(textEditSecretKey.text())

        textEditIdentityFile = widget.textEditIdentityFile
        controllerValues['identity file'] = str(textEditIdentityFile.text())

        if shouldCreateNew:
            automaton.addCloudControllerCredential(
                'Eucalyptus', 'euca2ools', **controllerValues)
            pass

        # save out to file
        configData = AppUtilsModule.createNewConfigDataObject()
        automaton.addConfigToSave(configData)
        AppUtilsModule.saveConfig(configData)

        return


    # END class Controller
    pass



from __future__ import with_statement

import copy
import logging
import os
import pickle
import shutil
import user
import uuid

import pypatterns.command as CommandPatternModule

import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule

import pomsets.builder as BuilderModule
import pomsets.command as ExecuteCommandModule
import pomsets.context as ContextModule
import pomsets.definition as PomsetDefinitionModule
import pomsets.library as DefinitionLibraryModule
import pomsets.error as ExecuteErrorModule
import pomsets.parameter as ParameterModule
import pomsets.python as PythonModule
import pomsets.resource as ResourceModule
import pomsets.task as TaskModule

import cloudpool.shell as ShellModule

import pomsets_app.controller.automaton as AutomatonModule



class ContextManager(ResourceModule.Struct):

    ATTRIBUTES = [
        'app',
        'automaton',
        'eventHandlers',
        'taskTable',
        'activeLibrary',
        'persistentLibrary',
        'transientLibrary',
        'pomsetBuilder',
        'resourcePath'
    ]

    COLUMNS_VIRTUALMACHINE_DISPLAY = [
        'Service', 'Instance ID', 'Image ID', 
        'IP', 'Key name', 'Enabled'
    ]

    def __init__(self):

        ResourceModule.Struct.__init__(self)

        self.eventHandlers([])

        table = RelationalModule.createTable(
            'tasks',
            ['task', 'pomset']
        )
        self.taskTable(table)

        self.activePomsetContext(None)

        self.pomsetBuilder(BuilderModule.Builder())

        # TODO:
        # generalize self._commands
        self._commands = {}

        return


    def commandPath(self, key, value=None):
        """
        this sets the path of a command
        """

        if value is not None:
            self._commands[key] = value
        return self._commands[key]


    def addActivePomsetContext(self, pomsetContext):
        self.activeLibrary().addPomsetContext(pomsetContext)
        self.activePomsetContext(pomsetContext)
        return

    def activePomset(self):
        activePomsetContext = self.activePomsetContext()
        if activePomsetContext is None:
            return None
        return activePomsetContext.pomset()
    
    def activePomsetContexts(self):
        filter = FilterModule.TRUE_FILTER
        activePomsetContexts = RelationalModule.Table.reduceRetrieve(
            self.activeLibrary().definitionTable(), filter, ['context'], [])
        return activePomsetContexts
    
    
    def activePomsetContext(self, value=None):
        if value is not None:
            self._activePomsetContext = value
        if not hasattr(self, '_activePomsetContext'):
            self._activePomsetContext = None
        return self._activePomsetContext
    
    
    def initializeAutomaton(self, automaton):

        self.automaton(automaton)
        automaton.contextManager(self)

        return


    """
    def initializeEventHandlers(self):
        ContextManagerModule.ContextManager.initializeEventHandlers(self)

        for key in ['canvas event handler',
                    'mouse event handler',
                    'key event handler']:
            
            eventHandler = self.app().getResourceValue(key)
            self.eventHandlers().append(eventHandler)
            eventHandler.contextManager = lambda: self
            pass

        return
    """


    def postEvent(self, event):


        for eventHandler in self.eventHandlers():
            eventHandler.AddPendingEvent(event)
            return

        return

    def createNewPomset(self, *args, **kwds):
        return self.pomsetBuilder().createNewNestPomset(*args, **kwds)


    #def createNewNode(self, **kwds):
    #    return self.pomsetBuilder().createNewNode(
    #        self.activePomset(), **kwds)

    def canConnect(self, port1, port2):
        port1Data = (port1.node.nodeData, port1.name)
        port2Data = (port2.node.nodeData, port2.name)
        return self.pomsetBuilder().canConnect(
            self.activePomset(),
            port1Data[0], port1Data[1],
            port2Data[0], port2Data[1])



    def connect(self, port1, port2):

        sourceNode = port1.node.nodeData
        sourceParameterId = port1.name
        targetNode = port2.node.nodeData
        targetParameterId = port2.name
        return self.pomsetBuilder().connect(
            self.activePomset(),
            sourceNode, sourceParameterId,
            targetNode, targetParameterId)


    def initializeLibraries(self):
        
        # TODO:
        # pull the bootstrapDefinitionDir from the config file instead

        appConfigDir = self.getAppConfigDir()
        bootstrapDefinitionDir = os.path.join(appConfigDir, 'library')
        
        # ensure that the pomset directories exists
        if not os.path.exists(bootstrapDefinitionDir):

            # TODO:
            # once we've migrated to 2.6
            # we can use the keyword "ignore"
            # to pass in a callable to copytree, 
            # so that we don't need to call rmtree

            defaultInitialLibrary = os.path.join(self.resourcePath(), 'library')
            # now copy the default initial library files over
            shutil.copytree(defaultInitialLibrary, bootstrapDefinitionDir)
            # remove the svn dir
            svnDir = os.path.sep.join([bootstrapDefinitionDir, '.svn'])
            if os.path.exists('.svn'):
                shutil.rmtree(svnDir)
            pass
        
        library = DefinitionLibraryModule.Library()
        library.bootstrapLoaderDefinitionsDir(bootstrapDefinitionDir)
            
        library.loadBootstrapLoaderDefinitions()
        self.persistentLibrary(library)
        
        library = DefinitionLibraryModule.Library()
        self.transientLibrary(library)

        library = DefinitionLibraryModule.Library()
        self.activeLibrary(library)

        return 

    def getContextForDefinition(self, definition, library=None):
        if library is None:
            library = self.transientLibrary()

        filter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdentityFilter(definition)
            )
        contexts = RelationalModule.Table.reduceRetrieve(
            library.definitionTable(),
            filter,
            ['context'],
            [])
        if len(contexts) is 0:
            raise ValueError('no context found for definition %s in library' % definition.name())
        if len(contexts) is not 1:
            raise ValueError('more than one context found for definition %s in library' % definition.name())
        return contexts[0]

        
    def getLoadedDefinitions(self):
        definitionTable = self.transientLibrary().definitionTable()
        for definitionRow in definitionTable.retrieve(
            columns=['context']):
            
            yield definitionRow[0]
        raise StopIteration
    
    
    def getLibraryDefinitions(self):
        
        definitionTable = self.persistentLibrary().definitionTable()
        for definitionRow in definitionTable.retrieve(
            columns=['context']):
            
            yield definitionRow[0]
        raise StopIteration
    

    def outputCommandsToStream(self, stream):
        # need to execute the pomset, 
        # with the exception that the execution
        # is actually to write to a stream
        # and that files should not be staged

        env = ExecuteCommandModule.PrintTaskCommand()
        env.prefix('# ')
        env.postfix('\n\n')

        env.outputStream(stream)


        envMap = {
            'shell process':env,
            'python eval':PythonModule.PythonEval()
            }

        # TODO;
        # if there are no threads allocated
        # then need to create a new thread
        # will also need to ensure 
        # that we remove the thread later

        # copy the active pomset
        # so that we don't screw with the active one
        # in case that it's already executing
        pomset = copy.copy(self.activePomset())
        reference = copy.copy(self.activePomsetContext().reference())
        reference.definitionToReference(pomset)
        pomset = reference

        commandBuilder = ExecuteCommandModule.CommandBuilder(
            ExecuteCommandModule.buildCommandFunction_commandlineArgs
        )
        commandBuilderMap = {
            'print task':commandBuilder,
            'shell process':commandBuilder,
            'python eval':PythonModule.CommandBuilder()
        }

        requestKwds = self.generateRequestKwds(
            executeEnvironmentMap = envMap,
            executeEnvironmentId = AutomatonModule.ID_EXECUTE_LOCAL,
            commandBuilderMap = commandBuilderMap)
        requestKwds['worker thread configures execute environment']=False


        self.executePomset(
            pomset = pomset,
            requestKwds=requestKwds,
            shouldWait = True)

        # remove the task for the pomset
        # from the task table
        # because we want to still be able to execute it
        # although it probably doesn't matter
        # because we're copying the pomset above anyway
        filter = RelationalModule.ColumnValueFilter(
            'pomset',
            FilterModule.IdentityFilter(pomset)
        )
        self.taskTable().removeRows(filter)

        return


    def _getTasksForPomset(self, pomset):
        filter = RelationalModule.ColumnValueFilter(
            'pomset',
            FilterModule.IdentityFilter(pomset)
        )
        tasks = RelationalModule.Table.reduceRetrieve(
            self.taskTable(),
            filter,
            ['task'],
            []
        )
        return tasks


    def hasTaskForRootPomset(self, rootPomset):
        tasks = self._getTasksForPomset(rootPomset)
        if len(tasks) is 0:
            return False
        return True

    def hasTaskForPomsetPath(self, pomsetReferencePath):
        if pomsetReferencePath is None:
            return False

        pomset = pomsetReferencePath[0]
        if not self.hasTaskForRootPomset(pomset):
            return False

        tasks = self._getTasksForPomset(pomset)
        parentTask = tasks[0]

        for pomsetReference in pomsetReferencePath[1:]:
            if not parentTask.hasChildTaskForDefinition(pomsetReference):
                return False
            childTask = parentTask.getChildTaskForDefinition(pomsetReference)
            parentTask = childTask
            pass

        return True



    def getTaskForPomsetPath(self, pomsetReferencePath):

        pomset = pomsetReferencePath[0]
        tasks = self._getTasksForPomset(pomset)

        if len(tasks) > 1:
            raise NotImplemented('not implemented for %s tasks' % len(tasks))

        parentTask = tasks[0]
        for pomsetReference in pomsetReferencePath[1:]:
            if not parentTask.hasChildTaskForDefinition(pomsetReference):
                return False
            childTask = parentTask.getChildTaskForDefinition(pomsetReference)
            parentTask = childTask
            pass
        return parentTask


    def getDefaultCommandBuilderMap(self):
        shellCommandBuilder = ExecuteCommandModule.CommandBuilder(
            ExecuteCommandModule.buildCommandFunction_commandlineArgs
            )
        commandBuilderMap = {
            'shell process':shellCommandBuilder,
            'python eval':PythonModule.CommandBuilder()
            }
        return commandBuilderMap


    def getDefaultExecuteEnvironmentMap(self):
        envMap = {
            'python eval':PythonModule.PythonEval()
            }
        return envMap


    def generateRequestKwds(self, 
                            executeEnvironmentMap=None,
                            executeEnvironmentId=None,
                            commandBuilderMap=None):


        if executeEnvironmentId is None and \
                'shell process' not in executeEnvironmentMap:
            raise NotImplementedError(
                'need to specify an execute environment or environment id')

        # make a copy of the defaultCommandBuilderMap
        # and overlay values from the map being passed in
        defaultMap = copy.copy(self.getDefaultCommandBuilderMap())
        if commandBuilderMap is not None:
            defaultMap.update(commandBuilderMap)
        commandBuilderMap = defaultMap

        kwds = {
            'command builder map':commandBuilderMap,
            'execute environment map':executeEnvironmentMap,
            'execute environment id':executeEnvironmentId,
            'worker thread configures execute environment':executeEnvironmentId is not None
            }
        return kwds


    def executePomset(self, pomset=None, 
                      requestKwds=None, shouldWait=False):

        # TODO:
        # first check to see if there's already a task
        # associated with the pomset to be executed
        if self.hasTaskForRootPomset(pomset):
            raise ExecuteErrorModule.ExecutionError(
                'there is already a task for the active pomset')

        # associate the task with the 
        # active pomset in the tasks table
        # need to do this before we execute
        # in case the execution errors
        task = TaskModule.CompositeTask()
        row = self.taskTable().addRow()
        row.setColumn('task', task)
        row.setColumn('pomset', pomset)


        self.automaton().executePomset(
            task=task,
            pomset=pomset, 
            requestKwds=requestKwds,
            shouldWait=shouldWait)


        return task


    def closeActivePomset(self):
        # TODO:
        # need to ensure that the task is no longer running

        self.closePomset(self.activePomsetContext())
        return
        
    
    def closePomset(self, pomsetContextToClose):

        # TODO:
        # similar to the add, this should be a command

        # remove from contextManager.taskTable
        filter = RelationalModule.ColumnValueFilter(
            'pomset',
            FilterModule.IdentityFilter(pomsetContextToClose.pomset())
        )
        self.taskTable().removeRows(filter)

        filter = RelationalModule.ColumnValueFilter(
            'context',
            FilterModule.IdentityFilter(pomsetContextToClose))
        self.activeLibrary().removeDefinition(filter)

        # TODO:
        # if there's another pomset open
        # then we should use that one
        # as the active pomset instead
        try:
            row = self.activeLibrary().definitionTable().retrieve(columns=['context']).next()
            self.activePomsetContext(row[0])
        except StopIteration:
            # there is no more active pomset context
            self._activePomsetContext = None

            pass

        return

    
    def imagePath(self):
        return os.path.join(self.resourcePath(), 'images')


    def getExecutableTypes(self):
        return ['Shell', 'Hadoop Jar', 'Hadoop Streaming', 'Hadoop Pipes', 'Python Eval']


    def getTypeForExecutable(self, executable):

        if executable.__class__ is ExecuteCommandModule.Executable:
            return 'Shell'

        import pomsets.hadoop as HadoopModule
        if executable.__class__ is HadoopModule.JarExecutable:
            # TODO:
            # need to inspect the jarFile
            jarFile = ' '.join(executable.jarFile())
            if jarFile == self.getHadoopStreamingJar():
                return 'Hadoop Streaming'

            return 'Hadoop Jar'

        if executable.__class__ is HadoopModule.PipesExecutable:
            return 'Hadoop Pipes'

        # TODO:
        # handle the case that the executable is a python function
        import pomsets.python as PythonModule
        if executable.__class__ is PythonModule.Function:
            return 'Python Eval'

        raise NotImplementedError(
            'not implemented for executable class %s' % executable.__class__)


    def getClassForExecutableType(self, executableType):
        if executableType in ['Shell']:
            return ExecuteCommandModule.Executable

        import pomsets.hadoop as HadoopModule
        if executableType in ['Hadoop Jar', 'Hadoop Streaming']:
            return HadoopModule.JarExecutable

        if executableType in ['Hadoop Pipes']:
            return HadoopModule.PipesExecutable

        if executableType in ['Python Eval']:
            import pomsets.python as PythonModule
            return PythonModule.Function

        raise NotImplementedError(
            'not implemented for executable type %s' % executableType)


    def getHadoopStreamingJar(self):
        configuration = self.automaton().hadoopConfigurations()['localhost']
        return configuration['streaming jar']


    # END class ContextManager
    pass



class QtContextManager(ContextManager):

    ATTRIBUTES = ContextManager.ATTRIBUTES + [
        'currentDisplayedPomsetInfo',
        'currentDisplayedPomsetIsEditable',
        'mainWindow'
        ]

    def getAppConfigDir(self):
        import user
        appConfigDir = os.path.join(user.home, '.pomsets')
        return appConfigDir

    def initializeEventHandlers(self):
        return


    # END class QtContextManager
    pass



import sys
import os
import user

import pomsets_app.utils as UtilsModule

def main(args):

    configDir = UtilsModule.getDefaultConfigDir()
    libraryDir = os.path.join(configDir, 'library')

    UtilsModule.backupLibrary(libraryDir)
    
    return


if __name__=="__main__":
    main(sys.argv)



import uuid


def createAtomicDefinition(contextManager, name=None):

    builder = contextManager.pomsetBuilder()
    executableObject = builder.createExecutableObject([], staticArgs=[])

    id = uuid.uuid4().hex

    # autogenerate the name of the definition
    if name is None:
        name = '_'.join(['definition', id[:3]])

    definitionContext = builder.createNewAtomicPomset(
        name=name, executableObject=executableObject)
    definition = definitionContext.pomset()
    definition.id(id)

    contextManager.transientLibrary().addPomsetContext(definitionContext)

    return definitionContext

